"""
Kafka Producer implementation class
"""

import json
import time

from kafka import KafkaProducer

KAFKA_BROKER_URL = "localhost:9092"
TOPIC_NAME = "metrics"


def send_message(producer, topic_name, message):
    """Send a message to the Kafka topic."""
    try:
        future = producer.send(topic_name, value=message)
        future.get(timeout=10)  # Ensure the message was sent successfully
        print(f"Message sent: {message}")
    except Exception as e:
        print(f"Failed to send message: {e}")


def get_kafka_producer(kafka_broker_url):
    """
    Initialize and return a Kafka producer.

    This function sets up a Kafka producer configured with retry and acknowledgment settings
    to ensure reliable message delivery. It serializes the messages to JSON format before
    sending to Kafka. The producer attempts to resend messages up to 5 times in case of
    failure, and waits for acknowledgment from all replicas to confirm message persistence.

    Parameters:
    - kafka_broker_url (str): The URL of the Kafka broker to connect to.

    Returns:
    - KafkaProducer: An instance of KafkaProducer configured for JSON message serialization.

    Kafka Configuration:
    - `bootstrap_servers`: Specifies the Kafka broker URL.
    - `retries`: Number of retries in case of message delivery failure (set to 5).
    - `acks`: Configured to 'all' to ensure the message is fully committed to the log by
      waiting for acknowledgment from all in-sync replicas.
    """
    kafka_config = {
        "bootstrap_servers": kafka_broker_url,
        "retries": 5,  # Retry sending messages on failure
        "acks": "all",  # Ensure message is committed to the log
    }

    return KafkaProducer(
        **kafka_config,
        value_serializer=lambda v: json.dumps(v).encode(
            "utf-8"
        ),  # Convert Python dict to JSON
    )


def send_kafka_message():
    """
    Sends a sequence of messages to a specified Kafka topic at regular intervals.

    This function initiates a Kafka producer, then sends messages to a Kafka topic
    every second in a loop, with each message containing a sequential number and a
    timestamp. The producer connection is closed after all messages are sent.

    Parameters:
    - None

    Returns:
    - None

    Notes:
    - This function is hardcoded to send messages with a 'number' field starting at 14 and incrementing up to 100,
      along with the current timestamp.
    - The Kafka producer is initialized with `KAFKA_BROKER_URL` and sends messages to `TOPIC_NAME`.
    """
    kafka_producer = get_kafka_producer(KAFKA_BROKER_URL)

    # Simulate sending messages every second
    for i in range(14, 100):
        message1 = {"number": i, "timestamp": time.time()}
        send_message(kafka_producer, TOPIC_NAME, message1)
        time.sleep(1)

    kafka_producer.close()


send_kafka_message()



"""
This module provides functionality related to Model actions via plugin.
"""

import json
import os
import subprocess
import sys
from datetime import datetime, timedelta
from typing import Any
import joblib
from kubernetes import client, config
from kubernetes.client.rest import ApiException
from .mlflowplugin import MlflowPlugin
from .. import plugin_config
from ..pluginmanager import PluginManager
from ..util import custom_serializer
from ..util import make_post_request, make_delete_request, make_get_request
from .kubeflowplugin import KubeflowPlugin


class NotebookPlugin:
    """
    Class for defining reusable components.
    """

    def __init__(self):
        """
        Initializes the ModelPlugin class.
        """
        self.section = "notebook_plugin"

    @staticmethod
    def link_model_to_dataset(dataset_id, model_id):
        """
        Links a model to a dataset using the provided API endpoint.

        This method sends a POST request to the API to associate a specified model
        with a given dataset. It uses the user's ID defined in the plugin configuration.

        Args:
            dataset_id (str): The ID of the dataset to link to the model.
            model_id (str): The ID of the model to be linked to the dataset.

        Returns:
            Response: The response object from the API call.

        Raises:
            requests.exceptions.RequestException: An error occurred when making the POST request.
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        data = {
            "model_id": model_id,
            "dataset_id": dataset_id,
        }
        # call the api
        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "link_dataset_model"
        )
        return make_post_request(url, data=data)

    def save_model_details_to_db(self, registered_model_name):
        """
        store model details in database
        :param registered_model_name: name of the registered model
        :return: id of model
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        data = {
            "name": registered_model_name,
            "version": self.get_model_latest_version(registered_model_name),
            "type": "sklearn",
            "description": f"{registered_model_name} model",
        }

        # call the api to register model
        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "models"
        )
        return make_post_request(url, data=data)

    @staticmethod
    def get_model_latest_version(registered_model_name: str):
        """
        return the latest version of registered model
        :param registered_model_name: model name to get the versions
        :return: latest version
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        latest_version_info = MlflowPlugin().search_model_versions(
            filter_string=f"name='{registered_model_name}'"
        )
        sorted_model_versions = sorted(
            latest_version_info, key=lambda x: int(x.version), reverse=True
        )

        if sorted_model_versions:
            latest_version = sorted_model_versions[0]
            return latest_version.version

        # print(f"No model versions found for {registered_model_name}")
        return 1

    @staticmethod
    def save_model_uri_to_db(model_id, model_uri):
        """
            method to call the api to save model uri
        :param model_id: model id of the model
        :param model_uri: model uri
        :return: API response
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        # call the api for saving model_uri
        data = {
            "file_type": plugin_config.FILE_TYPE,
            "model_id": model_id,
            "uri": model_uri,
            "description": f"model uri of model id :{model_id}",
        }
        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "models_uri"
        )
        return make_post_request(url, data=data)

    @staticmethod
    def delete_pipeline_details_from_db(pipeline_id):
        """
        delete the pipeline details
        :param pipeline_id: pipeline id
        :return:
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "pipeline"
        )
        return make_delete_request(url=url, path_params=pipeline_id)

    @staticmethod
    def list_runs_by_pipeline_id(pipeline_id):
        """
        list the pipeline run details
        :param pipeline_id: pipeline_id
        :return: list of run details
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "pipeline_runs"
        )
        response = make_get_request(url=url, path_params=pipeline_id)
        return response["data"]

    @staticmethod
    def delete_run_details_from_db(pipeline_id):
        """
         delete the pipeline details
        :param pipeline_id: pipeline_id
        :return: successful deletion message or 404 error
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "pipeline_runs"
        )
        return make_delete_request(url=url, path_params=pipeline_id)

    @staticmethod
    def get_pipeline_id_by_name(pipeline_name):
        """
        Retrieves the pipeline ID for a given pipeline name.

        Args:
            pipeline_name (str): The name of the pipeline to fetch the ID for.

        Returns:
            str: The ID of the specified pipeline if found.

        Raises:
            ValueError: If no pipeline with the given name is found.

        Example:
            pipeline_id = NotebookPlugin.get_pipeline_id_by_name("Example Pipeline")
        """
        kfp = KubeflowPlugin()
        pipelines_response = kfp.client().list_pipelines()
        pipeline_id = None
        if pipelines_response.pipelines:
            for pipeline in pipelines_response.pipelines:
                if pipeline.name == pipeline_name:
                    pipeline_id = pipeline.id
                    return pipeline_id

        if not pipeline_id:
            print(f"No pipeline found with the name '{pipeline_name}'")

    @staticmethod
    def list_pipelines_by_name(pipeline_name):
        """
        Lists all versions and runs of the specified pipeline by name.

        Args:
            pipeline_name (str): The name of the pipeline to fetch details for.

        Returns:
            dict: A dictionary containing the pipeline ID, versions,
             and runs of the specified pipeline.

        Raises:
            ValueError: If the pipeline name is invalid or not found.
            Exception: For any other issues encountered during the fetch operations.
        """
        # Fetch all versions of the specified pipeline
        kfp = KubeflowPlugin()

        pipeline_id = NotebookPlugin.get_pipeline_id_by_name(pipeline_name)
        versions_response = kfp.list_pipeline_versions(pipeline_id=pipeline_id)
        run_list = NotebookPlugin.list_runs_by_pipeline_id(pipeline_id=pipeline_id)
        result_dict = {
            "pipeline_id": pipeline_id,
            "versions": versions_response.versions,
            "runs": run_list,
        }
        return result_dict

    @staticmethod
    def save_pipeline_details_to_db(details):
        """
            save the details related to pipeline to the database
        :param details: dictionary with all the details of pipeline,run_details,task_details,experiments
        :return:
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        data = json.dumps(details, default=custom_serializer, indent=4)
        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "pipeline"
        )
        make_post_request(url=url, data=data)

    def log_model_by_model_file(self, model_file_path, model_name):
        """
            log_model in cogflow with the model_file
        :param model_file_path: file_path of model
        :param model_name: name of the model
        :return:
            data = {
                "artifact_uri" : 'artifact_uri of the model',
                "version" : "model version"
            }
        """
        model = self.load_pkl(model_file_path)
        mfp = MlflowPlugin()
        mfp.mlflow.set_experiment(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}_exp")
        with mfp.mlflow.start_run() as run:
            run_id = run.info.run_id
            mfp.mlflow.sklearn.log_model(
                model, "model", registered_model_name=model_name
            )
            print("Artifact_path", run.info.artifact_uri)
            print("run_id", run_id)
            print("model_name", model_name)
            latest_version = self.get_model_latest_version(model_name)
            data = {
                "artifact_uri": f"{run.info.artifact_uri}/model",
                "version": latest_version,
            }
            return data

    def install_and_import(self, package):
        """
            install and import the given package
        :param package: package to be installed
        :return:
        """
        try:
            __import__(package)
        except ImportError:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
            __import__(package)

    def load_pkl(self, file_path):
        """
            load the pkl file to joblib
        :param file_path: path of .pkl file
        :return:
        """
        try:
            with open(file_path, "rb") as file:
                return joblib.load(file)
        except ModuleNotFoundError as e:
            missing_module = str(e).split("'")[1]
            print(f"Module {missing_module} not found. Installing...")
            self.install_and_import(missing_module)
            print(f"Module {missing_module} installed. Trying to load the file again.")
            return self.load_pkl(file_path)

    def deploy_model(self, model_name, model_version, isvc_name):
        """

        :param model_name: name of the model
        :param model_version: version of the model
        :param isvc_name: service name to be created for the deployed model
        :return:
        """
        try:
            PluginManager().verify_activation(NotebookPlugin().section)
            mfp = MlflowPlugin()
            model_uri = mfp.get_model_uri(model_name=model_name, version=model_version)
            kfp = KubeflowPlugin()
            kfp.serve_model_v1(model_uri, name=isvc_name)
            return {
                "status": True,
                "msg": f"Model {model_name} deployed with service {isvc_name}",
            }
        except Exception as exp:
            raise exp

    @staticmethod
    def serialize_artifacts(artifacts):
        """
        Converts the artifacts dictionary into a JSON serializable format.
        Each artifact object is converted to its URI string representation.

        Args:
            artifacts (dict): The original artifacts dictionary.

        Returns:
            dict: A dictionary with JSON serializable artifact data.
        """
        serialized_artifacts = {}

        for key, artifact in artifacts.items():
            # Convert artifact objects (like ImageEvaluationArtifact) to their URI string representation
            if hasattr(artifact, "uri"):
                serialized_artifacts[key] = artifact.uri
            else:
                serialized_artifacts[key] = str(artifact)

        return {"validation_artifacts": serialized_artifacts}

    @staticmethod
    def model_recommender(model_name=None, classification_score=None):
        """
        Calls the model recommender API and returns the response.

        Args:
        - model_name (str): The name of the model to recommend (optional).
        - classification_score (list): A list of classification scores to consider(e.g., accuracy_score, f1_score,
         recall_score, log_loss, roc_auc, precision_score, example_count, score.). (optional).

        Returns:
        - dict: The response from the model recommender API.
        """
        # Verify plugin activation
        PluginManager().verify_activation(NotebookPlugin().section)

        PluginManager().load_config()

        # call the api for model recommend
        data = {"model_name": model_name, "classification_score": classification_score}
        url = os.getenv(plugin_config.API_BASEPATH) + PluginManager().load_path(
            "model_recommend"
        )
        return make_get_request(url, query_params=data)

    @staticmethod
    def get_pipeline_task_sequence_by_run_id(run_id):
        """
        Fetches the pipeline workflow and task sequence for a given run in Kubeflow.

        Args:
            run_id (str): The ID of the pipeline run to fetch details for.

        Returns:
            tuple: A tuple containing:
                - pipeline_workflow_name (str): The name of the pipeline's workflow (root node of the DAG).
                - task_structure (dict): A dictionary representing the task structure of the pipeline, with each node
                                         containing information such as task ID, pod name, status, inputs, outputs,
                                         and resource duration.

        The task structure contains the following fields for each node:
            - id (str): The unique ID of the task (node).
            - podName (str): The name of the pod associated with the task.
            - name (str): The display name of the task.
            - inputs (list): A list of input parameters for the task.
            - outputs (list): A list of outputs produced by the task.
            - status (str): The phase/status of the task (e.g., 'Succeeded', 'Failed').
            - startedAt (str): The timestamp when the task started.
            - finishedAt (str): The timestamp when the task finished.
            - resourcesDuration (dict): A dictionary representing the resources used (e.g., CPU, memory).
            - children (list): A list of child tasks (if any) in the DAG.

        Example:
            >>> run_id = "afcf98bb-a9af-4a34-a512-1236110150ae"
            >>> pipeline_name, task_structure = get_pipeline_task_sequence_by_run_id(run_id)
            >>> print(f"Pipeline Workflow Name: {pipeline_name}")
            >>> print("Task Structure:", task_structure)

        Raises:
            ValueError: If the root node (DAG) is not found in the pipeline.
        """

        # Initialize the Kubeflow client
        kfp_client = KubeflowPlugin().client()

        # Get the details of the specified run using the run ID
        run_details = kfp_client.get_run(run_id)

        # Parse the workflow manifest from the pipeline runtime
        workflow_graph = json.loads(run_details.pipeline_runtime.workflow_manifest)
        namespace = workflow_graph["metadata"]["namespace"]

        # Access the nodes in the pipeline graph
        nodes = workflow_graph["status"]["nodes"]

        # Initialize variables for the pipeline name and root node ID
        pipeline_workflow_name = None
        root_node_id = None

        # Iterate through nodes to find the DAG root (pipeline root node)
        for node_id, node_data in nodes.items():
            if node_data["type"] == "DAG":
                pipeline_workflow_name = node_data["displayName"]
                root_node_id = node_id
                break

        if not root_node_id:
            raise ValueError("Root DAG node not found in the pipeline run.")

        # Task structure to store the task details
        task_structure = {}

        # Recursive function to traverse the graph and build the task structure
        def traverse(node_id, parent=None):
            node = nodes[node_id]

            # Extract inputs, outputs, phase (status), and other information
            inputs = node.get("inputs", {}).get("parameters", [])
            outputs = node.get("outputs", [])
            phase = node.get("phase", "unknown")
            started_at = node.get("startedAt", "unknown")
            finished_at = node.get("finishedAt", "unknown")
            resources_duration = node.get("resourcesDuration", {})

            # Task information dictionary for the current node
            task_info = {
                "id": node_id,
                "namespace": namespace,
                "podName": node_id,  # Assuming podName is the same as node_id
                "name": node["displayName"],
                "inputs": inputs,
                "outputs": outputs,
                "status": phase,
                "startedAt": started_at,
                "finishedAt": finished_at,
                "resourcesDuration": resources_duration,
                "children": [],
            }

            # Add the task to the parent's children or to the task structure if it's the root
            if parent is None:
                task_structure[node_id] = task_info
            else:
                parent["children"].append(task_info)

            # Recursively traverse and process child nodes
            if "children" in node and node["children"]:
                for child_id in node["children"]:
                    traverse(child_id, task_info)

        # Begin traversal starting from the root node of the pipeline
        if root_node_id:
            traverse(root_node_id)

        response = {
            "run_id": run_id,
            "pipeline_workflow_name": pipeline_workflow_name,
            "namespace": namespace,
            "task_structure": task_structure,
        }

        return response

    @staticmethod
    def get_run_id_by_run_name(run_name):
        """
        Fetches the run_id of a pipeline run by its name, traversing all pages if necessary.

        Args:
            run_name (str): The name of the pipeline run to search for.

        Returns:
            str: The run_id if found, otherwise None.
        """
        next_page_token = None
        page_size = 100  # Set page size (adjust if needed)
        kfp_client = KubeflowPlugin().client()

        # Traverse through pages to find the matching run name
        while True:
            # Fetch the list of runs, providing the next_page_token to continue from the last point
            runs_list = kfp_client.list_runs(
                page_size=page_size, page_token=next_page_token
            )

            # Check the current page for the run with the specified name
            for run in runs_list.runs:
                if run.name == run_name:
                    return run.id

            # Check if there are more pages
            next_page_token = runs_list.next_page_token
            if not next_page_token:
                # No more pages, the run was not found
                break

        return None

    @staticmethod
    def get_pipeline_task_sequence_by_run_name(run_name):
        """
        Fetches the task structure of a pipeline run based on its name.

        Args:
            run_name (str): The name of the pipeline run to fetch task structure for.

        Returns:
            tuple: (pipeline_workflow_name, task_structure)
        Example:
            >>>run_name = "Run of test_pipeline (ad001)"
            >>>pipeline_name, task_structure = get_pipeline_task_sequence_by_run_name(run_name)
            >>>print(f'Pipeline Workflow Name: {pipeline_name}')
            >>>print("Task Structure:")
            >>>print(json.dumps(task_structure, indent=4))
        """
        kfp_client = KubeflowPlugin().client()

        # Fetch the run_id using the run_name
        run_id = NotebookPlugin().get_run_id_by_run_name(run_name)

        if not run_id:
            raise ValueError(f"No run found with name: {run_name}")

        # Get the details of the specified run by run_id
        run_details = kfp_client.get_run(run_id)

        # Parse the workflow manifest
        workflow_graph = json.loads(run_details.pipeline_runtime.workflow_manifest)
        namespace = workflow_graph["metadata"]["namespace"]

        # Access the nodes in the graph
        nodes = workflow_graph["status"]["nodes"]

        # Store the pipeline name and root node
        pipeline_workflow_name = None
        root_node_id = None

        for node_id, node_data in nodes.items():
            if node_data["type"] == "DAG":
                pipeline_workflow_name = node_data["displayName"]
                root_node_id = node_id
                break

        # Create a task representation structure
        task_structure = {}

        # Function to traverse the graph and build the task structure
        def traverse(node_id, parent=None):
            node = nodes[node_id]

            # Extract inputs, outputs, and additional information
            inputs = node.get("inputs", {}).get("parameters", [])
            outputs = node.get("outputs", [])
            phase = node.get("phase", "unknown")
            started_at = node.get("startedAt", "unknown")
            finished_at = node.get("finishedAt", "unknown")
            resources_duration = node.get("resourcesDuration", {})

            # Prepare the task information
            task_info = {
                "id": node_id,
                "namespace": namespace,
                "podName": node_id,
                "name": node["displayName"],
                "inputs": inputs,  # Include inputs
                "outputs": outputs,  # Include outputs
                "status": phase,
                "startedAt": started_at,
                "finishedAt": finished_at,
                "resourcesDuration": resources_duration,
                "children": [],
            }

            # Add task to the parent
            if parent is None:
                task_structure[node_id] = task_info
            else:
                parent["children"].append(task_info)

            # Recursively traverse child nodes
            if "children" in node and node["children"]:
                for child_id in node["children"]:
                    traverse(child_id, task_info)

        # Start traversing from the root node
        if root_node_id:
            traverse(root_node_id)

        response = {
            "run_id": run_id,
            "pipeline_workflow_name": pipeline_workflow_name,
            "namespace": namespace,
            "task_structure": task_structure,
        }

        return response

    @staticmethod
    def get_run_ids_by_pipeline_id(pipeline_id):
        """
        Fetches all run_ids for a given pipeline ID.

        Args:
            pipeline_id (str): The ID of the pipeline to search for.

        Returns:
            list: A list of run_ids for the matching pipeline ID.
        """
        run_ids = []
        next_page_token = None
        kfp_client = KubeflowPlugin().client()
        while True:
            runs_list = kfp_client.list_runs(page_size=100, page_token=next_page_token)
            for run in runs_list.runs:
                # Check if the run's pipeline_id matches the provided pipeline_id
                if run.pipeline_spec.pipeline_id == pipeline_id:
                    run_ids.append(run.id)

            # Check if there is a next page
            next_page_token = runs_list.next_page_token
            if not next_page_token:
                break  # Exit if there are no more pages

        return run_ids

    @staticmethod
    def get_pipeline_task_sequence_by_pipeline_id(pipeline_id):
        """
        Fetches the task structures of all pipeline runs based on the provided pipeline_id.

        Args:
            pipeline_id (str): The ID of the pipeline to fetch task structures for.

        Returns:
            list: A list of dictionaries containing pipeline workflow names and task structures for each run.
        Example:
            >>>pipeline_id = "1000537e-b101-4432-a779-768ec479c2b0"  # Replace with your actual pipeline_id
            >>>all_task_structures = get_pipeline_task_sequence_by_pipeline_id(pipeline_id)
            >>>for details in all_task_structures:
                >>>print(f'Run ID: {details["run_id"]}')
                >>>print(f'Pipeline Workflow Name: {details["pipeline_workflow_name"]}')
                >>>print("Task Structure:")
                >>>print(json.dumps(details["task_structure"], indent=4))
        """
        kfp_client = KubeflowPlugin().client()

        # Fetch all run_ids using the pipeline_id
        run_ids = NotebookPlugin().get_run_ids_by_pipeline_id(pipeline_id)

        if not run_ids:
            raise ValueError(f"No runs found for pipeline_id: {pipeline_id}")

        # Create a list to hold task structures for each run
        task_structures = []

        for run_id in run_ids:
            # Get the details of the specified run by run_id
            run_details = kfp_client.get_run(run_id)

            # Parse the workflow manifest
            workflow_graph = json.loads(run_details.pipeline_runtime.workflow_manifest)
            namespace = workflow_graph["metadata"]["namespace"]

            # Access the nodes in the graph
            nodes = workflow_graph["status"]["nodes"]

            # Store the pipeline name and root node
            pipeline_workflow_name = None
            root_node_id = None

            for node_id, node_data in nodes.items():
                if node_data["type"] == "DAG":
                    pipeline_workflow_name = node_data["displayName"]
                    root_node_id = node_id
                    break

            # Create a task representation structure
            task_structure = {}

            # Function to traverse the graph and build the task structure
            def traverse(node_id, parent=None):
                node = nodes[node_id]

                # Extract inputs, outputs, and additional information
                inputs = node.get("inputs", {}).get("parameters", [])
                outputs = node.get("outputs", [])
                phase = node.get("phase", "unknown")
                started_at = node.get("startedAt", "unknown")
                finished_at = node.get("finishedAt", "unknown")
                resources_duration = node.get("resourcesDuration", {})

                # Prepare the task information
                task_info = {
                    "id": node_id,
                    "namespace": namespace,
                    "podName": node_id,
                    "name": node["displayName"],
                    "inputs": inputs,  # Include inputs
                    "outputs": outputs,  # Include outputs
                    "status": phase,
                    "startedAt": started_at,
                    "finishedAt": finished_at,
                    "resourcesDuration": resources_duration,
                    "children": [],
                }

                # Add task to the parent
                if parent is None:
                    task_structure[node_id] = task_info
                else:
                    parent["children"].append(task_info)

                # Recursively traverse child nodes
                if "children" in node and node["children"]:
                    for child_id in node["children"]:
                        traverse(child_id, task_info)

            # Start traversing from the root node
            if root_node_id:
                traverse(root_node_id)

            # Append the task structure and workflow name for the current run_id
            task_structures.append(
                {
                    "run_id": run_id,
                    "pipeline_workflow_name": pipeline_workflow_name,
                    "namsepace": namespace,
                    "task_structure": task_structure,
                }
            )

        return task_structures

    @staticmethod
    def list_all_pipelines():
        """
        Lists all pipelines along with their IDs, handling pagination.

        Returns:
            list: A list of tuples containing (pipeline_name, pipeline_id).
        """
        kfp_client = KubeflowPlugin().client()

        pipelines_info = []
        next_page_token = None
        page_size = 100  # You can adjust this as needed

        while True:
            # Fetch all pipelines with pagination
            pipelines_list = kfp_client.list_pipelines(
                page_size=page_size, page_token=next_page_token
            )

            # Add the pipelines to the list
            for pipeline in pipelines_list.pipelines:
                pipelines_info.append((pipeline.name, pipeline.id))

            # Check if there is a next page
            next_page_token = pipelines_list.next_page_token
            if not next_page_token:
                break  # Exit the loop if there are no more pages

        return pipelines_info

    @staticmethod
    def get_run_ids_by_pipeline_name(pipeline_name):
        """
        Fetches all run_ids for a given pipeline name.

        Args:
            pipeline_name (str): The name of the pipeline to search for.

        Returns:
            list: A list of run_ids for the matching pipeline name.
        """
        run_ids = []
        next_page_token = None
        kfp_client = KubeflowPlugin().client()
        while True:
            runs_list = kfp_client.list_runs(page_size=100, page_token=next_page_token)
            for run in runs_list.runs:
                # Check if the run's pipeline name matches the provided pipeline name
                if run.pipeline_spec.pipeline_name == pipeline_name:
                    run_ids.append(run.id)

            # Check if there is a next page
            next_page_token = runs_list.next_page_token
            if not next_page_token:
                break  # Exit if there are no more pages

        return run_ids

    @staticmethod
    def get_all_run_ids():
        """
        Fetches all run_ids available in the system.

        Returns:
            list: A list of all run_ids.
        """
        run_ids = []
        next_page_token = None
        kfp_client = KubeflowPlugin().client()
        while True:
            runs_list = kfp_client.list_runs(page_size=100, page_token=next_page_token)
            for run in runs_list.runs:
                run_ids.append(run.id)

            next_page_token = runs_list.next_page_token
            if not next_page_token:
                break

        return run_ids

    @staticmethod
    def get_run_ids_by_name(run_name):
        """
        Fetches run_ids by run name.

        Args:
            run_name (str): The name of the run to search for.

        Returns:
            list: A list of run_ids matching the run_name.
        """
        run_ids = []
        next_page_token = None
        kfp_client = KubeflowPlugin().client()
        while True:
            runs_list = kfp_client.list_runs(page_size=100, page_token=next_page_token)
            for run in runs_list.runs:
                if run.name == run_name:
                    run_ids.append(run.id)

            next_page_token = runs_list.next_page_token
            if not next_page_token:
                break

        return run_ids

    @staticmethod
    def get_task_structure_by_task_id(task_id, run_id=None, run_name=None):
        """
        Fetches the task structure of a specific task ID, optionally filtered by run_id or run_name.

        Args:
            task_id (str): The task ID to look for.
            run_id (str, optional): The specific run ID to filter by. Defaults to None.
            run_name (str, optional): The specific run name to filter by. Defaults to None.

        Returns:
            list: A list of dictionaries containing run IDs and their corresponding task info if found.
        Example:
            >>>task_id = "test-pipeline-749dn-2534915009"
            >>>run_id = None  # "afcf98bb-a9af-4a34-a512-1236110150ae"
            >>>run_name = "Run of test_pipeline (ad001)"
            >>>get_task_structure_by_task_id(task_id, run_id, run_name)
        """
        kfp_client = KubeflowPlugin().client()

        # Fetch all run_ids available in the system
        run_ids = NotebookPlugin().get_all_run_ids()

        # If run_name is provided, filter by run_name
        if run_name:
            run_ids = NotebookPlugin().get_run_ids_by_name(run_name)

        # If run_id is provided, make it the only run to check
        if run_id:
            run_ids = [run_id] if run_id in run_ids else []

        task_structures = []

        for run_id in run_ids:
            # Get the details of the specified run by run_id
            run_details = kfp_client.get_run(run_id)

            # Parse the workflow manifest
            workflow_graph = json.loads(run_details.pipeline_runtime.workflow_manifest)
            namespace = workflow_graph["metadata"]["namespace"]

            # Access the nodes in the graph
            nodes = workflow_graph["status"]["nodes"]

            # Check if the task_id exists in the nodes
            if task_id in nodes:
                node_data = nodes[task_id]
                # Extract necessary details
                task_info = {
                    "id": task_id,
                    "namespace": namespace,
                    "name": node_data["displayName"],
                    "inputs": node_data.get("inputs", {}).get("parameters", []),
                    "outputs": node_data.get("outputs", []),
                    "status": node_data.get("phase", "unknown"),
                    "startedAt": node_data.get("startedAt", "unknown"),
                    "finishedAt": node_data.get("finishedAt", "unknown"),
                    "resourcesDuration": node_data.get("resourcesDuration", {}),
                    "run_id": run_id,
                }

                # Store the task info
                task_structures.append(task_info)
        if not task_structures:
            raise ValueError(f"No task found with ID: {task_id}.")
        return task_structures

    @staticmethod
    def load_k8s_config() -> None:
        """
        Loads the Kubernetes configuration.

        This method tries to load the in-cluster configuration if the code is running inside a pod.
        If it fails (e.g., if the code is running outside the cluster), it loads the kubeconfig file
        from the default location.

        Raises:
            config.config_exception.ConfigException: If the configuration could not be loaded.
        """
        try:
            # Load in-cluster config if running in a pod
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            # If not running in a pod, load the kubeconfig file
            config.load_kube_config()

    @staticmethod
    def get_pods(
        namespace=KubeflowPlugin().get_default_namespace(),
    ) -> Any:
        """
        Retrieves the list of pods in a specified namespace with a given label selector.

        Args:
            namespace (str): The namespace to list pods from.

        Returns:
            list: A list of V1Pod objects representing the pods that match the criteria.

        Raises:
            ApiException: If an error occurs while trying to list the pods.
        """
        NotebookPlugin().load_k8s_config()
        v_1 = client.CoreV1Api()
        try:
            pods = v_1.list_namespaced_pod(namespace=namespace)
            return pods
        except ApiException as exp:
            raise Exception(
                f"Exception when calling CoreV1Api->list_namespaced_pod: {str(exp)}"
            )

    @staticmethod
    def get_inference_service_logs(
        inference_service_name,
        namespace=KubeflowPlugin().get_default_namespace(),
        container_name="kserve-container",
    ) -> Any:
        """
        Retrieve logs for all pods matching the given label selector in a specified namespace.
        Args:
            namespace (str, optional): The Kubernetes namespace where the pods are located.
            inference_service_name (str): A inference_service to filter pods by their inference service.

            container_name (str, optional): The name of the container to fetch logs from.
                                            If not specified, logs are fetched from the default kserve-container.

        Returns:
            str: A JSON-formatted string containing the logs for each pod. Each log entry includes:
                 - "pod_name": The name of the pod.
                 - "namespace": The namespace of the pod.
                 - "logs": The logs for the pod, or a message indicating that no logs are available.
        """
        NotebookPlugin().load_k8s_config()

        pods = NotebookPlugin().get_pods(namespace)
        inference_pods = [
            pod for pod in pods.items if inference_service_name in pod.metadata.name
        ]
        log_entries = []
        for pod in inference_pods:
            pod_logs = NotebookPlugin().get_pod_logs(
                namespace, pod.metadata.name, container_name
            )
            log_entry = {
                "pod_name": pod.metadata.name,
                "namespace": pod.metadata.namespace,
                "logs": pod_logs if pod_logs else "No logs available.",
            }
            log_entries.append(log_entry)
        # Convert log entries list to JSON string
        json_logs = json.dumps(log_entries, indent=4)
        return {"logs": json_logs}

    @staticmethod
    def get_pod_logs(
        pod_name,
        namespace=KubeflowPlugin().get_default_namespace(),
        container_name=None,
    ):
        """
        Retrieves the logs of a specified pod.

        Args:
            namespace (str): The namespace of the pod.
            pod_name (str): The name of the pod.
            container_name (str, optional): The name of the container within the pod.
            If not specified, logs of the first container in the pod are returned.

        Returns:
            str: The logs of the specified pod.

        Raises:
            ApiException: If an error occurs while trying to retrieve the pod logs.
        """
        NotebookPlugin().load_k8s_config()

        v_1 = client.CoreV1Api()
        try:
            if container_name:
                raw_logs = v_1.read_namespaced_pod_log(
                    name=pod_name, namespace=namespace, container=container_name
                )
            elif "pipeline" in pod_name:
                container_name = "main"  # Container name is 'main' for pipeline pods
                raw_logs = v_1.read_namespaced_pod_log(
                    name=pod_name, namespace=namespace, container=container_name
                )
            elif "postgres" in pod_name:
                container_name = (
                    "postgres"  # Container name is 'main' for pipeline pods
                )
                raw_logs = v_1.read_namespaced_pod_log(
                    name=pod_name, namespace=namespace, container=container_name
                )
            elif "predictor" in pod_name:
                container_name = (
                    "kserve-container"  # Container name is 'main' for pipeline pods
                )
                raw_logs = v_1.read_namespaced_pod_log(
                    name=pod_name, namespace=namespace, container=container_name
                )

            else:
                raw_logs = v_1.read_namespaced_pod_log(
                    name=pod_name, namespace=namespace
                )

            parse_logs = [line.strip() for line in raw_logs.split("\n") if line.strip()]

            logs = json.dumps(parse_logs, indent=4)
            return logs
        except ApiException as exp:
            raise Exception(
                f"Exception when calling CoreV1Api->read_namespace_pod_log: {str(exp)}"
            )

    @staticmethod
    def get_run_ids_by_pipeline_workflow_name(pipeline_workflow_name):
        """
        Fetches all run IDs associated with a given pipeline workflow name.

        Args:
            pipeline_workflow_name (str): The workflow name of the pipeline.

        Returns:
            list: A list of run IDs associated with the provided workflow name.

        Example:
            >>> workflow_name = "pipeline-vzn7z"
            >>> run_ids = get_run_ids_by_pipeline_workflow_name(workflow_name)
            >>> print(run_ids)
        """
        kfp_client = KubeflowPlugin().client()

        # Initialize variables for pagination
        run_ids = []
        page_token = None

        while True:
            try:
                # Fetch a page of runs
                response = kfp_client.list_runs(page_token=page_token)
            except Exception as e:
                raise Exception(f"Error fetching runs from Kubeflow: {str(e)}")

            if not response or not response.runs:
                break

            # Process runs in the current page
            for run in response.runs:
                try:
                    # Fetch detailed run information using run ID
                    run_details = client.get_run(run.id)
                    # print(run_details)
                    # Extract workflow manifest from run details
                    workflow_manifest = json.loads(
                        run_details.pipeline_runtime.workflow_manifest
                    )

                    # Check if the workflow name matches the given pipeline_workflow_name
                    if pipeline_workflow_name in workflow_manifest["metadata"]["name"]:
                        # print(run.id)
                        run_ids.append(run.id)
                except Exception as e:
                    print(f"Error processing run {run.id}: {str(e)}")

            # Update the page token to fetch the next page
            page_token = response.next_page_token

            # Break the loop if there are no more pages
            if not page_token:
                break

        return run_ids

    @staticmethod
    def get_pipeline_task_sequence(pipeline_name=None, pipeline_workflow_name=None):
        """
        Fetches the task structures of all pipeline runs based on the provided pipeline name or pipeline workflow name.

        Args:
            pipeline_name (str, optional): The name of the pipeline to fetch task structures for.
            pipeline_workflow_name (str, optional): The workflow name of the pipeline to fetch task structures for.

        Returns:
            list: A list with details of task structures for each run.
        Example:
            >>> pipeline_workflow_name = "pipeline-vzn7z"
            >>> all_task_structures = get_pipeline_task_sequence(pipeline_workflow_name=pipeline_workflow_name)
            >>> for details in all_task_structures:
                    >>> print(f'Run ID: {details["run_id"]}')
                    >>> print(f'Pipeline Workflow Name: {details["pipeline_workflow_name"]}')
                    >>> print("Task Structure:")
                    >>> print(json.dumps(details["task_structure"], indent=4))

        Raises:
            ValueError: If neither pipeline_name nor pipeline_workflow_name is provided.
        """
        kfp_client = KubeflowPlugin().client()

        # Fetch all run_ids based on pipeline_name or pipeline_workflow_name
        if pipeline_name:
            run_ids = NotebookPlugin().get_run_ids_by_pipeline_name(pipeline_name)
        elif pipeline_workflow_name:
            run_ids = NotebookPlugin().get_run_ids_by_pipeline_workflow_name(
                pipeline_workflow_name
            )
        else:
            raise ValueError(
                "Either pipeline_name or pipeline_workflow_name must be provided."
            )

        if not run_ids:
            identifier = pipeline_name or pipeline_workflow_name
            raise ValueError(f"No runs found for identifier: {identifier}")

        # Create a dictionary to hold task structures for each run
        task_structures = {}
        output_details = []  # List to hold details to return

        for run_id in run_ids:
            # Get the details of the specified run by run_id
            run_details = kfp_client.get_run(run_id)

            # Parse the workflow manifest
            workflow_graph = json.loads(run_details.pipeline_runtime.workflow_manifest)
            namespace = workflow_graph["metadata"]["namespace"]

            # Access the nodes in the graph
            nodes = workflow_graph["status"]["nodes"]

            # Store the pipeline name and root node
            pipeline_workflow_name = None
            root_node_id = None

            for node_id, node_data in nodes.items():
                if node_data["type"] == "DAG":
                    pipeline_workflow_name = node_data["displayName"]
                    root_node_id = node_id
                    break

            # Create a task representation structure
            task_structure = {}

            # Function to traverse the graph and build the task structure
            def traverse(node_id, parent=None):
                node = nodes[node_id]

                # Extract inputs, outputs, and additional information
                inputs = node.get("inputs", {}).get("parameters", [])
                outputs = node.get("outputs", [])
                phase = node.get("phase", "unknown")
                started_at = node.get("startedAt", "unknown")
                finished_at = node.get("finishedAt", "unknown")
                resources_duration = node.get("resourcesDuration", {})

                # Prepare the task information
                task_info = {
                    "id": node_id,
                    "namespace": namespace,
                    "podName": node_id,
                    "name": node["displayName"],
                    "inputs": inputs,  # Include inputs
                    "outputs": outputs,  # Include outputs
                    "status": phase,
                    "startedAt": started_at,
                    "finishedAt": finished_at,
                    "resourcesDuration": resources_duration,
                    "children": [],
                }

                # Add task to the parent
                if parent is None:
                    task_structure[node_id] = task_info
                else:
                    parent["children"].append(task_info)

                # Recursively traverse child nodes
                if "children" in node and node["children"]:
                    for child_id in node["children"]:
                        traverse(child_id, task_info)

            # Start traversing from the root node
            if root_node_id:
                traverse(root_node_id)

            # Store the task structure for the current run_id
            task_structures[run_id] = {
                "pipeline_workflow_name": pipeline_workflow_name,
                "task_structure": task_structure,
            }

            # Append details to the output list
            output_details.append(
                {
                    "run_id": run_id,
                    "pipeline_workflow_name": pipeline_workflow_name,
                    "namespace": namespace,
                    "task_structure": task_structure,
                }
            )

        return output_details

    @staticmethod
    def get_pod_events(podname, namespace=KubeflowPlugin().get_default_namespace()):
        """
        Fetches Kubernetes events for a specific pod in a namespace.

        Args:
            podname (str): The name of the pod for which events are being fetched.
            namespace (str): The namespace of the pod.

        Returns:
            dict: A dictionary with event details related to the specified pod.
        """
        NotebookPlugin().load_k8s_config()

        # Initialize the CoreV1Api client
        v1 = client.CoreV1Api()
        event = {}
        try:
            # Fetch all events in the namespace
            events = v1.list_namespaced_event(namespace=namespace)

            event["kind"] = events.kind
            event["apiVersion"] = events.api_version
            event["metadata_resource_version"] = events.metadata.resource_version
            event["items"] = events.items
            return event
        except client.exceptions.ApiException as e:
            return {
                "error": f"Failed to fetch events: {str(e)}",
                "podname": podname,
                "namespace": namespace,
            }

    @staticmethod
    def convert_datetime(obj):
        """Recursively converts datetime objects to string."""
        if isinstance(obj, dict):
            return {k: NotebookPlugin().convert_datetime(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [NotebookPlugin().convert_datetime(i) for i in obj]
        elif isinstance(obj, datetime):
            return obj.isoformat()  # Convert datetime to string in ISO format
        else:
            return obj

    @staticmethod
    def get_pod_definition(podname, namespace=KubeflowPlugin().get_default_namespace()):
        """
        Fetches the pod definition for a specific pod in a specific namespace.

        Args:
            podname (str): The name of the pod for which the definition is being fetched.
            namespace (str): The namespace of the pod.

        Returns:
            dict: A dictionary with the full pod definition (specifications and metadata).
        """
        # Load Kubernetes configuration (use load_incluster_config() if running within a Kubernetes pod)
        NotebookPlugin().load_k8s_config()

        # Initialize the CoreV1Api client
        v1 = client.CoreV1Api()

        try:
            # Fetch the pod definition
            pod = v1.read_namespaced_pod(name=podname, namespace=namespace)

            # Convert the pod object to a dictionary and then recursively handle datetime objects
            pod_dict = pod.to_dict()
            pod_dict = NotebookPlugin().convert_datetime(
                pod_dict
            )  # Convert any datetime objects to strings
            pod_dict = json.dumps(pod_dict, indent=4)
            return pod_dict

        except client.exceptions.ApiException as e:
            return {
                "error": f"Failed to fetch pod definition: {str(e)}",
                "podname": podname,
                "namespace": namespace,
            }

    @staticmethod
    def get_deployments(namespace):
        """
        Fetches details of all InferenceServices in the given namespace and formats them.

        Args:
        - namespace (str): The Kubernetes namespace where InferenceServices are deployed. Defaults to "default".

        Returns:
        - list of dicts: A list of dictionaries with InferenceService details.
        """
        try:
            # Load Kubernetes configuration from the cluster (in-cluster configuration)
            config.load_incluster_config()  # Use this if running inside a cluster
            # config.load_kube_config()  # Uncomment this for local development with kubeconfig

            # Create the CustomObjectsApi to interact with KServe resources
            custom_api = client.CustomObjectsApi()

            # Fetch InferenceServices in the specified namespace
            kserve_deployments = custom_api.list_namespaced_custom_object(
                group="serving.kserve.io",  # KServe custom resources are in this API group
                version="v1beta1",  # KServe InferenceService typically uses v1beta1
                namespace=namespace,
                plural="inferenceservices",
            )

            inferenceservices_details = []

            # Iterate through the InferenceServices
            for isvc in kserve_deployments.get("items", []):
                # Extract relevant information from metadata and status
                name = isvc["metadata"]["name"]
                url = isvc["status"].get("url", "N/A")

                # Determine the ready status from conditions (if status is "True" for type "Ready")
                ready = (
                    "True"
                    if any(
                        condition.get("type") == "Ready"
                        and condition.get("status") == "True"
                        for condition in isvc["status"].get("conditions", [])
                    )
                    else "False"
                )

                # Get the previous, latest, and ready revisions
                latest_ready_revision = (
                    isvc["status"]
                    .get("components", {})
                    .get("predictor", {})
                    .get("latestReadyRevision", "N/A")
                )

                # Calculate the age of the InferenceService (from creationTimestamp)
                creation_timestamp = isvc["metadata"].get("creationTimestamp", None)
                if creation_timestamp:
                    creation_time = datetime.strptime(
                        creation_timestamp, "%Y-%m-%dT%H:%M:%SZ"
                    )
                    age = str(datetime.now() - creation_time).split(".", maxsplit=1)[0]

                else:
                    age = "Unknown"

                # Collect all the details into a dictionary
                inferenceservices_details.append(
                    {
                        "NAME": name,
                        "URL": url,
                        "READY": ready,
                        "LATESTREADYREVISION": latest_ready_revision,
                        "AGE": age,
                    }
                )

            return inferenceservices_details

        except ApiException as e:
            print(f"Exception when calling API: {e}")
            return []

    @staticmethod
    def calculate_duration(start_time, end_time):
        """
        Calculate the duration between two datetime objects.

        Args:
            start_time (datetime): The start time of the run.
            end_time (datetime): The end time of the run.

        Returns:
            str: Duration in the format 'HH:MM:SS'. If either start_time or end_time
                 is None, returns "N/A".
        """
        if start_time and end_time:
            duration = end_time - start_time
            return str(timedelta(seconds=duration.total_seconds()))
        return "N/A"

    @staticmethod
    def parse_runs(runs_response):
        """
        Parse the list of Kubeflow Pipeline (KFP) runs into a structured format.

        Args:
            runs_response (kfp_server_api.models.ApiListRunsResponse): The response object from
                the KFP `list_runs` API containing the runs information.

        Returns:
            list[dict]: A list of dictionaries where each dictionary contains the following:
                - "run_name" (str): The name of the run.
                - "run_id" (str): The unique ID of the run.
                - "status" (str): The current status of the run (e.g., "Succeeded", "Failed").
                - "duration" (str): The duration of the run in 'HH:MM:SS' format or "N/A".
                - "experiment_id" (str): The ID of the experiment associated with the run.
                - "start_time" (str): The start time of the run in ISO 8601 format or "N/A".
        """
        parsed_runs = []
        for run in runs_response.runs:
            run_name = run.name
            run_id = run.id
            start_time = run.created_at.isoformat() if run.created_at else "N/A"
            experiment_id = None
            for ref in run.resource_references:
                if ref.key.type == "EXPERIMENT":
                    experiment_id = ref.key.id
                    break
            status = run.status
            duration = NotebookPlugin().calculate_duration(
                run.created_at, run.finished_at
            )
            parsed_runs.append(
                {
                    "run_name": run_name,
                    "run_id": run_id,
                    "status": status,
                    "duration": duration,
                    "experiment_id": experiment_id,
                    "start_time": start_time,
                }
            )
        return parsed_runs

    @staticmethod
    def list_all_kfp_runs():
        """
        List all Kubeflow Pipeline (KFP) runs by iterating through all pages of results.

        This method retrieves and parses all available KFP runs using the KFP client API,
        handling pagination via `next_page_token`.

        Returns:
            list[dict]: A list of parsed runs, where each run is represented as a dictionary
                        containing:
                - "run_name" (str): The name of the run.
                - "run_id" (str): The unique ID of the run.
                - "status" (str): The current status of the run (e.g., "Succeeded", "Failed").
                - "duration" (str): The duration of the run in 'HH:MM:SS' format or "N/A".
                - "experiment_id" (str): The ID of the experiment associated with the run.
                - "start_time" (str): The start time of the run in ISO 8601 format or "N/A".
        """
        parsed_runs = []
        next_page_token = None
        kfp_client = KubeflowPlugin().client()
        while True:
            runs_response = kfp_client.list_runs(page_token=next_page_token)
            # print(runs_response)# Fetch KFP runs
            if runs_response and runs_response.runs:
                parsed_runs.extend(NotebookPlugin().parse_runs(runs_response))
            next_page_token = runs_response.next_page_token
            if not next_page_token:  # Stop when there are no more pages
                break

        return parsed_runs



"""
knative eventing plugin
"""

import os
import re
from kubernetes import client, dynamic
from kubernetes.dynamic import DynamicClient
from kubernetes.client.exceptions import ApiException
from kubernetes.client import api_client, ApiClient
from kubernetes.client import (
    V1ObjectMeta,
    V1Deployment,
    V1DeploymentSpec,
    V1LabelSelector,
    V1PodTemplateSpec,
    V1PodSpec,
    V1Container,
    V1VolumeMount,
    V1Volume,
    V1ConfigMapVolumeSource,
    V1ConfigMap,
)
from .. import util, plugin_config
from .kubeflowplugin import KubeflowPlugin


class KnativePlugin:
    """
    Class for defining reusable components.
    """

    def __init__(self):
        """
        Initializes the KnativePlugin class.
        """
        self.section = "knative_plugin"

    @staticmethod
    def check_dataset_exists(api_url, dataset_name, query_params=None):
        """
        Checks if a dataset with the exact name exists in the DB.

        :param api_url: URL to the dataset API endpoint.
        :param dataset_name: Name of the dataset to look for (exact match).
        :param query_params: Optional additional query parameters.
        :return: Boolean indicating existence, and matched dataset(s) if any.
        """
        if query_params is None:
            query_params = {"limit": 10}
        datasets = util.make_get_request(
            api_url, query_params=query_params, paginate=True
        )

        matching = [item for item in datasets if item["dataset_name"] == dataset_name]

        if matching:
            return True, matching
        else:
            print(f"No dataset found with name: {dataset_name}")
            return False, []

    @staticmethod
    def get_broker_and_topic_by_dataset_id(api_url, dataset_id, query_params=None):
        """
        Fetch broker_name and topic_name if the dataset_id matches.

        :param api_url: API endpoint that returns dataset message details.
        :param dataset_id: ID of the dataset to validate.
        :param query_params: Optional query parameters.
        :return: Tuple (exists_flag, broker_name, topic_name)
        """
        try:
            path_params = f"{dataset_id}/message/details"

            response = util.make_get_request(
                api_url, path_params=path_params, query_params=query_params
            )

            if not isinstance(response, dict):
                print("Unexpected response format")
                return False, None, None, None

            data = response.get("data", {})

            dataset = data.get("dataset")
            broker = data.get("broker_details")
            topic = data.get("topic_details")

            if dataset and dataset.get("id") == dataset_id:
                broker_name = broker.get("broker_name") if broker else None
                broker_port = broker.get("broker_port") if broker else None
                topic_name = topic.get("topic_name") if topic else None
                return True, broker_name, broker_port, topic_name
            else:
                print(f"No matching dataset ID: {dataset_id}")
                return False, None, None, None

        except Exception as e:
            print(f"Error occurred: {e}")
            return False, None, None, None

    @staticmethod
    def deploy_kafka_sink(name: str, namespace: str, topic: str, bootstrap_server: str):
        """
        Deploy or patch a KafkaSink in the given namespace.
        Args:
            name (str): KafkaSink name
            namespace (str): Kubernetes namespace
            topic (str): Kafka topic
            bootstrap_server (str): Kafka bootstrap server (e.g. kafka:9092)
        """

        KubeflowPlugin().load_k8s_config()
        k8s_client = client.ApiClient()
        dyn_client = DynamicClient(k8s_client)

        kafka_sink_manifest = {
            "apiVersion": "eventing.knative.dev/v1alpha1",
            "kind": "KafkaSink",
            "metadata": {"name": name, "namespace": namespace},
            "spec": {
                "topic": topic,
                "bootstrapServers": [bootstrap_server],  # list
                "contentMode": "binary",
            },
        }
        try:
            kafka_sink_resource = dyn_client.resources.get(
                api_version="eventing.knative.dev/v1alpha1", kind="KafkaSink"
            )

            # Try to get the KafkaSink
            try:
                kafka_sink_resource.get(name=name, namespace=namespace)

                print(f"KafkaSink '{name}' exists. Patching...")

                kafka_sink_resource.patch(
                    body=kafka_sink_manifest,
                    name=name,
                    namespace=namespace,
                    content_type="application/merge-patch+json",
                )

            except ApiException as e:
                if e.status == 404:

                    print(f"KafkaSink '{name}' not found. Creating...")

                    kafka_sink_resource.create(
                        body=kafka_sink_manifest, namespace=namespace
                    )
                else:
                    raise

            print(
                f"KafkaSink '{name}' successfully deployed in namespace '{namespace}'."
            )

        except Exception as e:
            print(f"Failed to deploy KafkaSink: {e}")

    @staticmethod
    def deploy_kafka_source(
        namespace,
        bootstrap_server,
        topic,
        source_name=None,
        consumer_group=None,
        sequence_name=None,
    ):
        """
        Deploy a Knative KafkaSource to consume messages from a Kafka topic
        and route them to a Knative Sequence.

        This function uses the dynamic Kubernetes client to create a
        KafkaSource custom resource that listens to a specified Kafka topic
        and sends messages to a Sequence sink in the same namespace.

        Parameters:
            namespace (str): Kubernetes namespace where the KafkaSource and Sequence exist.
            bootstrap_server (str): Kafka bootstrap server address (e.g., "kafka-cluster-kafka-bootstrap.kafka:9092").
            topic (str): Kafka topic name to subscribe to.
            source_name (str, optional): Name for the KafkaSource resource. Defaults to None.
            consumer_group (str, optional): Kafka consumer group ID. Defaults to None.
            sequence_name (str, optional): Name of the Knative Sequence to use as the sink.
            Defaults to None.

        Returns:
            None

        Raises:
            Prints errors if creation fails or if resource already exists.
        """
        # Load kubeconfig
        KubeflowPlugin().load_k8s_config()

        # Dynamic client
        dyn_client = dynamic.DynamicClient(api_client.ApiClient())

        # Load KafkaSource CRD resource
        kafka_source_api = dyn_client.resources.get(
            api_version="sources.knative.dev/v1beta1", kind="KafkaSource"
        )

        # Define KafkaSource manifest
        kafka_source_manifest = {
            "apiVersion": "sources.knative.dev/v1beta1",
            "kind": "KafkaSource",
            "metadata": {"name": source_name, "namespace": namespace},
            "spec": {
                "bootstrapServers": [bootstrap_server],
                "consumerGroup": consumer_group,
                "topics": [topic],
                "sink": {
                    "ref": {
                        "apiVersion": "flows.knative.dev/v1",
                        "kind": "Sequence",
                        "name": sequence_name,
                        "namespace": namespace,
                    }
                },
            },
        }

        try:
            # Create KafkaSource
            kafka_source_api.create(body=kafka_source_manifest, namespace=namespace)
            print(f"KafkaSource '{source_name}' created in namespace '{namespace}'")
        except Exception as e:
            if "AlreadyExists" in str(e):
                print(
                    f"KafkaSource '{source_name}' already exists. You may need to delete and recreate "
                    f"it if updating immutable fields."
                )
            else:
                print(f"Failed to create KafkaSource: {e}")

    @staticmethod
    def deploy_sequence(sequence_name, namespace, model_isvc_name, kafka_sink_name):
        """
        Deploy a Knative Sequence that sends data to a model and replies to a KafkaSink.

        Args:
            sequence_name (str): Name for the Knative Sequence.
            namespace (str): Kubernetes namespace.
            model_isvc_name (str): Name of the KServe model (InferenceService).
            kafka_sink_name (str): Name of the KafkaSink to route replies to.
        """
        try:
            # Load kubeconfig (in-cluster or local)
            KubeflowPlugin().load_k8s_config()

            # Create dynamic client from loaded config
            k8s_client = ApiClient()
            dyn_client = DynamicClient(k8s_client)

            # Define Sequence manifest
            sequence_manifest = {
                "apiVersion": "flows.knative.dev/v1",
                "kind": "Sequence",
                "metadata": {
                    "name": sequence_name,
                    "namespace": namespace,
                },
                "spec": {
                    "channelTemplate": {
                        "apiVersion": "messaging.knative.dev/v1",
                        "kind": "InMemoryChannel",
                    },
                    "steps": [
                        {
                            "uri": f"http://{model_isvc_name}-predictor.{namespace}.svc.cluster.local/v1/models"
                            f"/{model_isvc_name}:predict"
                        }
                    ],
                    "reply": {
                        "ref": {
                            "apiVersion": "eventing.knative.dev/v1alpha1",
                            "kind": "KafkaSink",
                            "name": kafka_sink_name,
                            "namespace": namespace,
                        }
                    },
                },
            }

            # Deploy
            sequence_api = dyn_client.resources.get(
                api_version="flows.knative.dev/v1", kind="Sequence"
            )

            try:
                # Check if it exists
                sequence_api.get(name=sequence_name, namespace=namespace)
                print(f"Sequence '{sequence_name}' exists. Patching...")
                response = sequence_api.patch(
                    name=sequence_name,
                    namespace=namespace,
                    body=sequence_manifest,
                    content_type="application/merge-patch+json",
                )
            except ApiException as e:
                if e.status == 404:
                    print(f"Creating Sequence '{sequence_name}'...")
                    response = sequence_api.create(
                        body=sequence_manifest, namespace=namespace
                    )
                else:
                    raise

            print(
                f"Sequence '{response.metadata.name}' deployed in namespace '{namespace}'."
            )

        except ApiException as e:
            print(f"Failed to deploy Sequence: {e.status}")
            print("Reason:", e.reason)
            print("Details:", e.body)

    @staticmethod
    def create_configmap_nats_kafka_bridge(
        namespace: str, configmap_name: str, config_data: str
    ):
        """
        Create or update a Kubernetes ConfigMap for the NATS-Kafka bridge configuration.

        If the ConfigMap already exists, this method merges new connector entries into
        the existing 'connectors.conf' while preventing duplication. If it doesn't exist,
        a new ConfigMap is created.

        Parameters:
            namespace (str): Kubernetes namespace for the ConfigMap.
            configmap_name (str): Name of the ConfigMap.
            config_data (str): YAML-like bridge config (as a string) to include.

        Returns:
            None

        Raises:
            kubernetes.client.rest.ApiException: For Kubernetes API errors other than 404.
        """
        v1 = client.CoreV1Api()
        key = "connectors.conf"

        try:
            existing_configmap = v1.read_namespaced_config_map(
                name=configmap_name, namespace=namespace
            )
            existing_data = existing_configmap.data.get(key, "")

            # Extract connect arrays from both existing and new config strings
            existing_connectors = re.findall(r"{[^{}]+}", existing_data)
            new_connectors = re.findall(r"{[^{}]+}", config_data)

            # Merge unique entries (basic string-level match)
            merged_connectors = existing_connectors.copy()
            for connector in new_connectors:
                if connector not in existing_connectors:
                    merged_connectors.append(connector)

            # Replace the connect block
            common_preamble = re.split(
                r"connect\s*:\s*\[", existing_data or config_data
            )[0].strip()
            merged_config = (
                f"{common_preamble}\nconnect: [\n  "
                + ",\n  ".join(merged_connectors)
                + "\n]"
            )

            print(f"Updating ConfigMap '{configmap_name}' with merged connectors...")
            existing_configmap.data[key] = merged_config
            v1.replace_namespaced_config_map(
                name=configmap_name, namespace=namespace, body=existing_configmap
            )

        except ApiException as e:
            if e.status == 404:
                print(f"Creating ConfigMap '{configmap_name}'...")
                config_map = V1ConfigMap(
                    metadata=V1ObjectMeta(name=configmap_name, namespace=namespace),
                    data={key: config_data},
                )
                v1.create_namespaced_config_map(namespace=namespace, body=config_map)
            else:
                raise

    @staticmethod
    def deploy_nats_kafka_bridge_deployment(
        namespace: str, deployment_name: str, configmap_name: str
    ):
        """
        Creates or updates a Kubernetes Deployment for the NATS-Kafka bridge.

        This function defines a Deployment resource for running the `natsio/nats-kafka` bridge
        container, which connects NATS subjects to Kafka topics based on configuration from a
        specified ConfigMap. It mounts the configuration file from the ConfigMap at
        `/etc/bridge/connectors.conf`.

        If the Deployment already exists in the given namespace, it will be updated.
        Otherwise, a new Deployment will be created.

        Parameters:
            namespace (str): The Kubernetes namespace where the Deployment should be created or updated.
            deployment_name (str): The name of the Deployment.
            configmap_name (str): The name of the ConfigMap containing the bridge configuration.

        Returns:
            None

        Raises:
            kubernetes.client.rest.ApiException: If the Kubernetes API call fails for reasons
            other than the Deployment not being found (404).
        """
        apps_api = client.AppsV1Api()
        labels = {"app": deployment_name}

        deployment = V1Deployment(
            metadata=V1ObjectMeta(name=deployment_name, namespace=namespace),
            spec=V1DeploymentSpec(
                replicas=1,
                selector=V1LabelSelector(match_labels=labels),
                template=V1PodTemplateSpec(
                    metadata=V1ObjectMeta(labels=labels),
                    spec=V1PodSpec(
                        containers=[
                            V1Container(
                                name="bridge",
                                image="natsio/nats-kafka:latest",
                                args=["-c", "/etc/bridge/connectors.conf"],
                                volume_mounts=[
                                    V1VolumeMount(
                                        name="config", mount_path="/etc/bridge"
                                    )
                                ],
                            )
                        ],
                        volumes=[
                            V1Volume(
                                name="config",
                                config_map=V1ConfigMapVolumeSource(name=configmap_name),
                            )
                        ],
                    ),
                ),
            ),
        )

        try:
            apps_api.read_namespaced_deployment(
                name=deployment_name, namespace=namespace
            )
            print(f"Updating Deployment '{deployment_name}'...")
            apps_api.replace_namespaced_deployment(
                name=deployment_name, namespace=namespace, body=deployment
            )
            print(
                f"Deployment '{deployment_name}' updated successfully in namespace '{namespace}'."
            )
        except ApiException as e:
            if e.status == 404:
                print(f"Creating Deployment '{deployment_name}'...")
                apps_api.create_namespaced_deployment(
                    namespace=namespace, body=deployment
                )
                print(
                    f"Deployment '{deployment_name}' created successfully in namespace '{namespace}'."
                )
            else:
                raise

    @staticmethod
    def connect(source_dataset, model_isvc, destination_dataset):
        """
        normally in cogflow you check each dataset ,
        1) all of them should be streaming type
        2) then create sink and source and sequence for them
        3) for source and destination if the type is nats , create bridge for each of them as well
        """

        base_url = os.getenv(plugin_config.API_BASEPATH)
        dataset_url = f"{base_url}{plugin_config.MESSAGE_BROKER_DATASETS_URL}"
        namespace = KubeflowPlugin().get_default_namespace()

        # check if source_dataset exists
        source_exists, source_matched = KnativePlugin().check_dataset_exists(
            dataset_url, source_dataset
        )

        # check if destination_dataset exists
        destination_exists, destination_matched = KnativePlugin().check_dataset_exists(
            dataset_url, destination_dataset
        )

        # Ensure both exist
        if not source_exists or not destination_exists:
            return

        # Extract the first matched dataset info
        source_info = source_matched[0]
        dest_info = destination_matched[0]

        if source_info["data_source_type"] not in [10, 11] or dest_info[
            "data_source_type"
        ] not in [10, 11]:
            print("Both datasets must be of streaming type.")
            return
        # Check if both have same data_source_type
        elif source_info["data_source_type"] != dest_info["data_source_type"]:
            print("Source and destination datasets have different data_source_type.")
            return
        # If both are valid streaming datasets
        else:
            # Check for broker and topic details
            (
                source_exists,
                source_broker_name,
                source_broker_port,
                source_topic,
            ) = KnativePlugin().get_broker_and_topic_by_dataset_id(
                api_url=dataset_url,
                dataset_id=source_info["id"],
                query_params={"limit": 10},
            )

            (
                dest_exists,
                dest_broker_name,
                dest_broker_port,
                dest_topic,
            ) = KnativePlugin().get_broker_and_topic_by_dataset_id(
                api_url=dataset_url,
                dataset_id=dest_info["id"],
                query_params={"limit": 10},
            )
            if (
                source_info["data_source_type"] == 11
                and dest_info["data_source_type"] == 11
            ):

                KnativePlugin().create_configmap_nats_kafka_bridge(
                    namespace=namespace,
                    configmap_name="nats-kafka-config",
                    config_data=plugin_config.NATS_KAFKA_CONNECTOR_JSON,
                )

                KnativePlugin().deploy_nats_kafka_bridge_deployment(
                    namespace=namespace,
                    deployment_name="nats-kafka-bridge",
                    configmap_name="nats-kafka-config",
                )
            elif source_info["data_source_type"] in [10, 11] and dest_info[
                "data_source_type"
            ] in [10, 11]:

                if source_exists and dest_exists:
                    try:
                        # adapt with source_dataset
                        KnativePlugin().deploy_kafka_source(
                            namespace=namespace,
                            bootstrap_server=f"{source_broker_name}:{source_broker_port}",
                            topic=f"{source_topic}",
                            source_name=f"{source_dataset}-events-source-seq",
                            consumer_group=f"{source_dataset}-kserve-group",
                            sequence_name=f"{dest_topic}-sequence",
                        )
                        # adapt with dest_dataset
                        KnativePlugin().deploy_kafka_sink(
                            name=f"{dest_topic}-sink",
                            namespace=namespace,
                            topic=f"{dest_topic}",
                            bootstrap_server=f"{dest_broker_name}:{dest_broker_port}",
                        )
                        KnativePlugin().deploy_sequence(
                            sequence_name=f"{dest_topic}-sequence",
                            namespace=namespace,
                            model_isvc_name=model_isvc,
                            kafka_sink_name=f"{dest_topic}-sink",
                        )
                    except Exception as e:
                        print(f"Error occurred during deployment: {e}")

                else:
                    print("No matching dataset or invalid response")



"""
This module provides functionality related to Dataset upload via plugin.
"""

import io
import os
from typing import Union
from urllib.parse import urlparse
import numpy as np
import pandas as pd
import requests
from minio import Minio
from mlflow.models.signature import ModelSignature
from scipy.sparse import csr_matrix, csc_matrix
from .. import plugin_config
from ..pluginmanager import PluginManager
from ..util import make_post_request, make_get_request
from .notebook_plugin import NotebookPlugin
from .mlflowplugin import MlflowPlugin
from .kubeflowplugin import KubeflowPlugin


class DatasetMetadata:
    """
    Class used for  metadata of Dataset
    """

    def __init__(self, name, description, file_path, dataset_type: str):
        self.name = name
        self.description = description
        self.file_path = file_path
        self.dataset_type = dataset_type

    def is_file_path(self):
        """
            method to check if the source of  dataset
            is local file path
        :return: boolean true or false
        """
        return os.path.isfile(self.file_path)

    def is_external_url(self):
        """
            method to check if source of dataset is
            external url
        :return: boolean true or false
        """
        parsed_url = urlparse(self.file_path)
        return bool(parsed_url.scheme) and parsed_url.netloc

    def to_dict(self):
        """
        return  object as dictionary
        """
        return {
            "name": self.name,
            "description": self.description,
            "file_path": self.file_path,
            "dataset_type": self.dataset_type,
        }


class DatasetPlugin:
    """
    A class to handle dataset-related operations.
    Attributes:
        None
    """

    def __init__(self):
        """
        Initializes DatasetPlugin with environment variables.
        """
        # Retrieve MinIO connection details from environment variables
        self.minio_endpoint = os.getenv(plugin_config.MINIO_ENDPOINT_URL)
        # Check if the environment variable exists and has a value
        if self.minio_endpoint:
            # Remove the http:// or https:// prefix using string manipulation
            if self.minio_endpoint.startswith(("http://", "https://")):
                # Find the index where the protocol ends
                protocol_end_index = self.minio_endpoint.find("//") + 2
                # Get the remaining part of the URL (without the protocol)
                self.minio_endpoint = self.minio_endpoint[protocol_end_index:]
        else:
            print("MLFLOW_S3_ENDPOINT_URL environment variable is not set.")
        self.minio_access_key = os.getenv(plugin_config.MINIO_ACCESS_KEY)
        self.minio_secret_key = os.getenv(plugin_config.MINIO_SECRET_ACCESS_KEY)
        self.section = "dataset_plugin"

    def create_minio_client(self):
        """
        Creates a MinIO client object.
        Returns:
            Minio: The MinIO client object.
        """
        # Verify plugin activation
        PluginManager().verify_activation(self.section)
        return Minio(
            self.minio_endpoint,
            access_key=self.minio_access_key,
            secret_key=self.minio_secret_key,
            secure=False,
        )  # Change to True if using HTTPS

    def query_endpoint_and_download_file(self, url, output_file, bucket_name):
        """
        Queries an endpoint and downloads a file from it.
        Args:
            url (str): The URL of the endpoint.
            output_file (str): The name of the output file to save.
            bucket_name (str): The name of the bucket.
        Returns:
            tuple: A tuple containing a boolean indicating success and the file URL if successful.
        """
        # Verify plugin activation
        PluginManager().verify_activation(self.section)
        try:
            response = requests.get(url, timeout=10)
            if response.status_code == 200:
                self.save_to_minio(response.content, output_file, bucket_name)
                return True
            print(f"Request failed with status code {response.status_code}")
            raise Exception("Request could not be successful due to error")
        except requests.exceptions.RequestException as exp:
            print(f"An error occurred: {exp}")
            raise Exception("Exception occurred during the requested operation")

    def save_to_minio(self, file_content, output_file, bucket_name):
        """
        Saves a file to MinIO.
        Args:
            file_content (bytes): The content of the file to be uploaded.
            output_file (str): The name of the file to be uploaded.
            bucket_name (str): The name of the bucket to upload the file to.
        Returns:
            str: The presigned URL of the uploaded file.
        """
        # Verify plugin activation
        PluginManager().verify_activation(self.section)
        # Initialize MinIO client
        minio_client = self.create_minio_client()
        object_name = output_file
        # Check if the bucket exists, if not, create it
        bucket_exists = minio_client.bucket_exists(bucket_name)
        if not bucket_exists:
            try:
                minio_client.make_bucket(bucket_name)
                print(f"Bucket '{bucket_name}' created successfully.")
            except Exception as exp:
                print(f"Bucket '{bucket_name}' couldnot be created.")
                raise exp
        # Put file to MinIO
        try:
            # Upload content to MinIO bucket
            minio_client.put_object(
                bucket_name,
                object_name,
                io.BytesIO(file_content),
                len(file_content),
            )
            print(
                f"File {output_file} uploaded successfully to MinIO bucket"
                f" {bucket_name} as {object_name}."
            )
            presigned_url = minio_client.presigned_get_object(bucket_name, object_name)
            print(f"Access URL for '{object_name}': {presigned_url}")
            return presigned_url
        except Exception as err:
            print(f"Error uploading file: {err}")
            raise Exception(f"Error uploading file: {err}")

    def delete_from_minio(self, object_name, bucket_name):
        """
        Deletes a file from MinIO.
        Args:
            object_name (str): The name of the object (file) to be deleted.
            bucket_name (str): The name of the bucket containing the file.
        Returns:
            bool: True if the file was successfully deleted, False otherwise.
        """
        # Verify plugin activation
        PluginManager().verify_activation(self.section)
        # Initialize MinIO client
        minio_client = self.create_minio_client()
        try:
            # Check if the object exists
            object_exists = minio_client.stat_object(bucket_name, object_name)
            if object_exists:
                # Delete the object from the bucket
                minio_client.remove_object(bucket_name, object_name)
                print(
                    f"File '{object_name}' deleted successfully from bucket '{bucket_name}'."
                )
                return True
            print(f"File '{object_name}' does not exist in bucket '{bucket_name}'.")
            return False
        except Exception as err:
            print(
                f"Error deleting file '{object_name}' from bucket '{bucket_name}': {err}"
            )
            return False

    @staticmethod
    def register_dataset(
        dataset_type: int, name: str, file_path: str, description: str = None
    ):
        """
        Register a dataset by uploading a file to the API using make_post_request.

        :param dataset_type: 0 (train), 1 (inference), 2 (both)
        :param name: Dataset name
        :param description: Optional dataset description
        :param file_path: Path to the dataset file
        :return: API response in JSON
        """
        url = f"{os.getenv('API_BASEPATH')}/datasets/file"

        # Form fields for multipart/form-data
        form_data = {
            "dataset_type": str(dataset_type),
            "name": name,
            "description": description or "",
        }

        # Header with kubeflow user id
        headers = {
            "kubeflow-userid": KubeflowPlugin().get_current_user_from_namespace()
        }

        # Files dictionary; key must match FastAPI parameter name "files"
        with open(file_path, "rb") as f:
            files = {"files": (os.path.basename(file_path), f)}
            response = make_post_request(
                url=url, data=form_data, files=files, headers=headers
            )

        return response

    def save_dataset_details(self, dataset_metadata):
        """
        Save dataset details by registering the dataset via the API.

        Parameters
        ----------
        dataset_metadata : DatasetMetadata
            Instance containing name, description, file_path, and dataset_type.

        Returns
        -------
        str
            The dataset_id returned by the API.
        """
        # Verify plugin activation
        PluginManager().verify_activation(self.section)

        response = self.register_dataset(
            dataset_type=int(dataset_metadata.dataset_type),
            name=dataset_metadata.name,
            file_path=dataset_metadata.file_path,
            description=dataset_metadata.description,
        )
        dataset_id = response["data"]["dataset_id"]
        return dataset_id

    def log_model_with_dataset(
        self,
        sk_model,
        artifact_path,
        dataset: DatasetMetadata,
        conda_env=None,
        code_paths=None,
        serialization_format="cloudpickle",
        registered_model_name=None,
        signature: ModelSignature = None,
        input_example: Union[
            pd.DataFrame,
            np.ndarray,
            dict,
            list,
            csr_matrix,
            csc_matrix,
            str,
            bytes,
            tuple,
        ] = None,
        await_registration_for=300,
        pip_requirements=None,
        extra_pip_requirements=None,
        pyfunc_predict_fn="predict",
        metadata=None,
    ):
        """
        Log a scikit-learn model to Mlflow and link dataset to model.

        Args:
            sk_model: The scikit-learn model to be logged.
            artifact_path (str): The run-relative artifact path to which the model artifacts will
            be saved.
            conda_env (str, optional): The path to a Conda environment YAML file. Defaults to None.
            code_paths (list, optional): A list of local filesystem paths to Python files that
            contain code to be
            included as part of the model's logged artifacts. Defaults to None.
            dataset (DatasetMetadata): Metadata of the dataset to link with the model.
            serialization_format (str, optional): The format used to serialize the model. Defaults
            to "cloudpickle".
            registered_model_name (str, optional): The name under which to register the model with
            Mlflow. Defaults to None.
            signature (ModelSignature, optional): The signature defining model input and output
            data types and shapes. Defaults to None.
            input_example (Union[pd.DataFrame, np.ndarray, dict, list, csr_matrix, csc_matrix, str,
            bytes, tuple], optional): An example input to the model. Defaults to None.
            await_registration_for (int, optional): The duration, in seconds, to wait for the
            model version to finish being created and is in the READY status. Defaults to 300.
            pip_requirements (str, optional): A file in pip requirements format specifying
            additional pip dependencies for the model environment. Defaults to None.
            extra_pip_requirements (str, optional): A string containing additional pip dependencies
            that should be added to the environment. Defaults to None.
            pyfunc_predict_fn (str, optional): The name of the function to invoke for prediction,
            when the model is a PyFunc model. Defaults to "predict".
            metadata (dict, optional): A dictionary of metadata to log with the model.
            Defaults to None.

        Returns:
            Model: The logged scikit-learn model.

        Raises:
            Exception: If an error occurs during the logging process.

        """
        # Verify plugin activation
        PluginManager().verify_activation(self.section)

        result = MlflowPlugin().log_model(
            sk_model=sk_model,
            artifact_path=artifact_path,
            conda_env=conda_env,
            code_paths=code_paths,
            serialization_format=serialization_format,
            registered_model_name=registered_model_name,
            signature=signature,
            input_example=input_example,
            await_registration_for=await_registration_for,
            pip_requirements=pip_requirements,
            extra_pip_requirements=extra_pip_requirements,
            pyfunc_predict_fn=pyfunc_predict_fn,
            metadata=metadata,
        )
        # save model details in DB
        response = NotebookPlugin().save_model_details_to_db(registered_model_name)
        model_id = response["data"]["id"]
        # save the dataset details
        dataset_id = self.save_dataset_details(dataset)
        # link model and dataset
        NotebookPlugin().link_model_to_dataset(dataset_id, model_id)
        return result

    @staticmethod
    def get_dataset(dataset_id: int, endpoint: str):
        """
        Generic method to call dataset API endpoints like /datasets/prometheus/{id}.

        :param dataset_id: Dataset ID to fetch
        :param endpoint: API endpoint path (e.g., "/datasets/prometheus")
        :return: API JSON response
        """
        PluginManager().load_config()

        url = f"{os.getenv(plugin_config.API_BASEPATH)}{endpoint}"

        headers = {
            "kubeflow-userid": KubeflowPlugin().get_current_user_from_namespace()
        }

        resp = make_get_request(
            url=url,
            path_params=dataset_id,
            headers=headers,
        )

        return resp.get("data")



"""
This module contains unit tests for the functions of the DatasetPlugin and related functionality.
"""

import unittest
from unittest.mock import patch, MagicMock

import minio
import requests

from ..cogflow.plugins.dataset_plugin import DatasetPlugin
from ..cogflow import (
    query_endpoint_and_download_file,
    save_to_minio,
    delete_from_minio,
)


class TestDatasetPlugin(unittest.TestCase):
    """
    This class contains test cases for the DatasetPlugin functionalities like
    querying endpoints, saving to and deleting from Minio, and dataset registration.
    """

    @patch("requests.get")
    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.save_to_minio")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_query_endpoint_and_download_file_success(
        self, mock_plugin_activation, mock_save_to_minio, mock_requests_get
    ):
        """
        Test successful file download from the dataset endpoint.
        Verifies if file content is saved to Minio and plugin activation is checked.
        """
        # Arrange
        url = "http://dataset.com/dataset"
        output_file = "dataset.csv"
        bucket_name = "mlpipeline"
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b"Test dataset content"
        mock_requests_get.return_value = mock_response

        # mock_minio_instance = MagicMock()
        mock_save_to_minio.return_value = "http://dataset.com/dataset.csv"

        # Act
        success = query_endpoint_and_download_file(
            url=url, output_file=output_file, bucket_name=bucket_name
        )

        # Assert
        self.assertTrue(success)
        mock_save_to_minio.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("requests.get")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_query_endpoint_and_download_file_failure(
        self, mock_plugin_activation, mock_requests_get
    ):
        """
        test for query_endpoint_and_download_file
        """
        # Arrange
        dataset_plugin = DatasetPlugin()
        url = "http://dataset.com/dataset"
        bucket_name = "mlpipeline"
        output_file = "dataset.csv"
        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_requests_get.return_value = mock_response

        # Act & Assert
        with self.assertRaises(Exception):
            dataset_plugin.query_endpoint_and_download_file(
                url, output_file, bucket_name
            )
        mock_plugin_activation.assert_called_once()

    @patch("requests.get")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_request_exception_query_endpoint(
        self, mock_plugin_activation, mock_requests_get
    ):
        """
        test for verify_activation
        """
        # Arrange
        dataset_plugin = DatasetPlugin()
        url = "http://dataset.com/dataset"
        bucket_name = "mlpipeline"
        output_file = "dataset.csv"
        # mock_response = MagicMock()
        # mock_response.status_code = 404
        # mock_requests_get.return_value = mock_response
        mock_requests_get.side_effect = requests.exceptions.RequestException(
            "Request failed"
        )
        with self.assertRaises(Exception):
            dataset_plugin.query_endpoint_and_download_file(
                url, output_file, bucket_name
            )
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("os.getenv")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_save_to_minio_success(
        self, mock_plugin_activation, mock_getenv, mock_create_minio_client
    ):
        """
        test for create_minio_client
        """

        # Arrange
        mock_minio_client = MagicMock()
        mock_create_minio_client.return_value = mock_minio_client
        mock_getenv.side_effect = lambda x: {
            "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
            "AWS_ACCESS_KEY_ID": "minio",
            "AWS_SECRET_ACCESS_KEY": "minio123",
        }[x]

        file_content = b"Test dataset content"
        output_file = "dataset.csv"

        mock_minio_instance = MagicMock()
        mock_minio_instance.bucket_exists.return_value = True
        mock_minio_instance.presigned_get_object.return_value = (
            "http://dataset.com/dataset.csv"
        )

        # Act
        save_to_minio(file_content, output_file, bucket_name="mlpipeline")
        # Assertions
        mock_minio_client.bucket_exists.assert_called_once_with("mlpipeline")
        mock_minio_client.presigned_get_object.assert_called_once()
        mock_minio_client.put_object.assert_called_once()
        mock_create_minio_client.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("os.getenv")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_minio_bucket_creation(
        self,
        mock_plugin_activation,
        mock_getenv,
        mock_create_minio_client,
    ):
        """
        test for create_minio_client
        """
        # Arrange
        mock_minio_client = MagicMock()
        mock_create_minio_client.return_value = mock_minio_client
        mock_getenv.side_effect = lambda x: {
            "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
            "AWS_ACCESS_KEY_ID": "minio",
            "AWS_SECRET_ACCESS_KEY": "minio123",
        }[x]

        dataset_plugin = DatasetPlugin()
        file_content = b"Test dataset content"
        output_file = "dataset.csv"

        mock_minio_client.bucket_exists.return_value = False
        dataset_plugin.save_to_minio(
            file_content, output_file, bucket_name="bucket_name"
        )
        mock_minio_client.make_bucket.assert_called_once()
        mock_minio_client.make_bucket.assert_called_once_with("bucket_name")
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("os.getenv")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_minio_bucket_creation_exception(
        self,
        mock_plugin_activation,
        mock_getenv,
        mock_create_minio_client,
    ):
        """
        test for create_minio_client
        """
        # Arrange
        mock_minio_client = MagicMock()
        mock_create_minio_client.return_value = mock_minio_client
        mock_getenv.side_effect = lambda x: {
            "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
            "AWS_ACCESS_KEY_ID": "minio",
            "AWS_SECRET_ACCESS_KEY": "minio123",
        }[x]

        dataset_plugin = DatasetPlugin()
        file_content = b"Test dataset content"
        output_file = "dataset.csv"

        mock_minio_client.bucket_exists.return_value = False
        mock_minio_client.make_bucket.side_effect = Exception(
            "Bucket Cannot be created"
        )
        with self.assertRaises(Exception):
            dataset_plugin.save_to_minio(
                file_content, output_file, bucket_name="bucket_name"
            )
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("os.getenv")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_save_to_exception(
        self, mock_plugin_activation, mock_getenv, mock_create_minio_client
    ):
        """
        test for create_minio_client
        """
        # Arrange
        mock_minio_client = MagicMock()
        mock_create_minio_client.return_value = mock_minio_client
        mock_getenv.side_effect = lambda x: {
            "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
            "AWS_ACCESS_KEY_ID": "minio",
            "AWS_SECRET_ACCESS_KEY": "minio123",
        }[x]

        dataset_plugin = DatasetPlugin()
        file_content = b"Test dataset content"
        output_file = "dataset.csv"

        # mock_minio_client.bucket_exists.return_value = False
        # mock_minio_client.make_bucket.side_effect = Exception("Bucket Cannot be created")
        mock_minio_client.put_object.side_effect = Exception(
            "Error while storing object"
        )
        with self.assertRaises(Exception):
            dataset_plugin.save_to_minio(
                file_content, output_file, bucket_name="bucket_name"
            )
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_delete_from_minio_success(
        self, mock_plugin_activation, mock_create_minio_client
    ):
        """
        test for delete_from_minio
        """
        # Arrange
        object_name = "test_object"
        bucket_name = "test_bucket"

        mock_minio_instance = MagicMock()
        mock_create_minio_client.return_value = mock_minio_instance
        mock_minio_instance.stat_object.return_value = True  # Object exists

        # Act
        result = delete_from_minio(object_name, bucket_name)

        # Assert
        self.assertTrue(result)
        mock_minio_instance.remove_object.assert_called_once_with(
            bucket_name, object_name
        )
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_delete_from_minio_object_not_found(
        self, mock_plugin_activation, mock_create_minio_client
    ):
        """
        test for delete_from_minio
        """
        # Arrange
        dataset_plugin = DatasetPlugin()
        object_name = "test_object"
        bucket_name = "test_bucket"

        mock_minio_instance = MagicMock()
        mock_create_minio_client.return_value = mock_minio_instance
        mock_minio_instance.stat_object.return_value = False  # Object does not exist

        # Act
        result = dataset_plugin.delete_from_minio(object_name, bucket_name)

        # Assert
        self.assertFalse(result)
        mock_minio_instance.remove_object.assert_not_called()
        mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.plugins.dataset_plugin.DatasetPlugin.create_minio_client")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_delete_from_minio_exception(
        self, mock_plugin_activation, mock_create_minio_client
    ):
        """
        test for delete_from_minio
        """
        # Arrange
        dataset_plugin = DatasetPlugin()
        object_name = "test_object"
        bucket_name = "test_bucket"

        mock_minio_instance = MagicMock()
        mock_create_minio_client.return_value = mock_minio_instance
        mock_minio_instance.stat_object.side_effect = Exception("Test error")

        # Act
        result = dataset_plugin.delete_from_minio(object_name, bucket_name)

        # Assert
        self.assertFalse(result)
        mock_minio_instance.remove_object.assert_not_called()
        mock_plugin_activation.assert_called_once()

    @patch("os.getenv")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_create_minio_client(self, mock_plugin_activation, mock_getenv):
        """
        test for create_minio_client
        """
        # Define test parameters
        mock_getenv.side_effect = lambda x: {
            "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
            "AWS_ACCESS_KEY_ID": "minio",
            "AWS_SECRET_ACCESS_KEY": "minio123",
        }[x]

        # Create the MinioClient object
        dataset_plugin = DatasetPlugin()
        minio_client = dataset_plugin.create_minio_client()

        # Assert that the MinioClient object is created correctly
        self.assertIsInstance(minio_client, minio.api.Minio)
        mock_plugin_activation.assert_called_once()

    # def test_log_model_with_dataset(self):
    #     """
    #     test for log_model_with_dataset
    #     """
    #     with patch("mlflow.sklearn.log_model") as mock_log_model:
    #         with patch(
    #             "cogflow.cogflow.plugins.notebook_plugin.NotebookPlugin.get_model_latest_version"
    #         ) as mock_model_version:
    #             with patch("mlflow.active_run") as mock_active_run:
    #                 with patch("os.getenv") as mock_env:
    #                     with patch("requests.post") as mock_requests_post:
    #                         # Create a mock run object
    #                         mock_run = MagicMock()
    #                         mock_run.info.run_id = "12345"
    #
    #                         # Set the return value of mlflow.active_run()
    #                         mock_active_run.return_value = mock_run
    #                         mock_env.side_effect = lambda x: {
    #                             "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
    #                             "AWS_ACCESS_KEY_ID": "minio",
    #                             "AWS_SECRET_ACCESS_KEY": "minio123",
    #                             "API_BASEPATH": "http://randomn",
    #                             "TIMER_IN_SEC": "10",
    #                             "MLFLOW_TRACKING_URI": "http://mlflow",
    #                             "ML_TOOL": "ml_flow",
    #                             "FILE_TYPE": "0",
    #                         }[x]
    #
    #                         mock_response = {
    #                             "data": {"dataset_id": 5, "id": 1},
    #                             "message": "Dataset linked with model successfully",
    #                         }
    #                         mock_requests_post.return_value.status_code = 201
    #                         mock_requests_post.return_value.json.return_value = (
    #                             mock_response
    #                         )
    #
    #                         # Mock model
    #                         sk_model = MagicMock()
    #                         artifact_path = "model"
    #                         registered_model_name = "testmodel"
    #
    #                         # Dataset details
    #                         source = (
    #                             "https://archive.ics.uci.edu/static/public/17"
    #                             "/breast+cancer+wisconsin+diagnostic.zip"
    #                         )
    #                         file_format = "zip"
    #                         name = "breast+cancer+wisconsin+diagnostic.zip"
    #                         description = "Breast cancer wisconsin diagnotic dataset"
    #
    #                         dm = DatasetMetadata(name, description, source, file_format)
    #                         mock_model_version.return_value = 1
    #                         # Define any other necessary inputs for the log_model method
    #
    #                         # Call the method under test
    #                         log_model_with_dataset(
    #                             sk_model=sk_model,
    #                             artifact_path=artifact_path,
    #                             registered_model_name=registered_model_name,
    #                             dataset=dm,
    #                         )
    #                         mock_log_model.assert_called_once()
    #
    # @patch("requests.post")
    # @patch("os.getenv")
    # def test_register_dataset(self, mock_env, mock_requests_post):
    #     """
    #     test for register_dataset
    #     """
    #     mock_env.side_effect = lambda x: {
    #         "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
    #         "AWS_ACCESS_KEY_ID": "minio",
    #         "AWS_SECRET_ACCESS_KEY": "minio123",
    #         "API_BASEPATH": "http://randomn",
    #         "TIMER_IN_SEC": "10",
    #         "MLFLOW_TRACKING_URI": "http://mlflow_server",
    #         "ML_TOOL": "mlflow",
    #         "FILE_TYPE": "0",
    #     }[x]
    #
    #     mock_response = {
    #         "data": {"dataset_id": 5, "id": 1},
    #         "message": "Dataset uploaded successfully",
    #     }
    #     mock_requests_post.return_value.status_code = 201
    #     mock_requests_post.return_value.json.return_value = mock_response
    #     # Dataset details
    #     source = (
    #         "https://archive.ics.uci.edu/static/public/17"
    #         "/breast+cancer+wisconsin+diagnostic.zip"
    #     )
    #     file_format = "zip"
    #     name = "breast+cancer+wisconsin+diagnostic.zip"
    #     description = "Breast cancer wisconsin diagnotic dataset"
    #
    #     dm = DatasetMetadata(name, description, source, file_format)
    #     result = register_dataset(dm)
    #     self.assertEqual(result, mock_response)


if __name__ == "__main__":
    unittest.main()



"""
This module defines constants for configuring Mlflow and AWS S3, as well as database credentials
 and MinIO settings.

Attributes:
    TRACKING_URI (str): The URI of the Mlflow tracking server. Used to specify where Mlflow logs and
    artifacts should be stored.
    S3_ENDPOINT_URL (str): The endpoint URL for the AWS S3 service. Needed for accessing and
     storing data in an S3 bucket.
    ACCESS_KEY_ID (str): The access key ID for AWS S3 authentication. Used in conjunction with
     SECRET_ACCESS_KEY for secure access to S3.
    SECRET_ACCESS_KEY (str): The secret access key for AWS S3 authentication. Paired with
     ACCESS_KEY_ID for secure access to S3.
    TIMER_IN_SEC (int): The interval in seconds for operations that require timing, such as
     periodic checks or updates. Default is set to 10 seconds.
    ML_TOOL (str): The name of the machine learning tool. Currently set to "mlflow" to
     specify the use of the Mlflow framework.
    MINIO_ENDPOINT_URL (str): The endpoint URL for the MinIO service, an alternative to AWS S3.
     Used for accessing and storing data.
    MINIO_ACCESS_KEY (str): The access key for MinIO authentication. Used with
     MINIO_SECRET_ACCESS_KEY for secure access to MinIO.
    MINIO_SECRET_ACCESS_KEY (str): The secret access key for MinIO authentication. Combined with
     MINIO_ACCESS_KEY for secure access.
     nats_kafka_connector_json (str): nats_kafka_connector_json
"""

COGFLOW_CONFIG_FILE_PATH = "COGFLOW_CONFIG_FILE_PATH"

TRACKING_URI = "MLFLOW_TRACKING_URI"
S3_ENDPOINT_URL = "MLFLOW_S3_ENDPOINT_URL"
ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"
TIMER_IN_SEC = 10
ML_TOOL = "mlflow"
BUCKET_NAME = "mlflow"
MODEL_TYPE = "sklearn"

MINIO_ENDPOINT_URL = "MLFLOW_S3_ENDPOINT_URL"
MINIO_ACCESS_KEY = "AWS_ACCESS_KEY_ID"
MINIO_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"

SQLALCHEMY_TEST_DATABASE_URI = "sqlite:///test.db"
TESTING_CONFIG = "config.app_config.TestingConfig"

FILE_TYPE = 0

API_BASEPATH = "API_BASEPATH"
TIME_OUT = 300

# log_model method parameters
SERIALIZATION_FORMAT = "cloudpickle"
AWAIT_REGISTRATION_FOR = 300
PYFUNC_PREDICT_FN = "predict"

# evaluate method parameters
ENV_MANAGER = "local"

# search_registered_models method parameters
MAX_RESULTS = 100

# create_fl_component_from_func method parameters
CONTAINER_PORT = 8080

# DEX configuration
POD_NAME = "dex-auth-0"
NAMESPACE = "kubeflow"
CONTAINER = "dex"
CONFIG_PATH = "/etc/dex/config.docker.yaml"
GRPC_PORT = 5557


# Docker images
BASE_IMAGE = "hiroregistry/cogflow:latest"
FL_COGFLOW_BASE_IMAGE = "hiroregistry/flcogflow:latest"

# MESSAGE_BROKER plugin
MESSAGE_BROKER_DATASETS_URL = "/datasets"
MESSAGE_BROKER_DATASETS_REGISTER = "/broker/register"
MESSAGE_BROKER_TOPIC_REGISTER = "/topic/register"
MESSAGE_BROKER_TOPIC_DATASETS_REGISTER = "/message/register"
MESSAGE_BROKER_TOPIC_DATASETS_DETAILS = "/message/details"

# KNATIVE PLUGIN
NATS_KAFKA_CONNECTOR_JSON = """
reconnectinterval: 5000,

connecttimeout: 5000,

logging: {
  time: true,
  debug: false,
  trace: false,
  colors: true,
  pid: false,
}

monitoring: {
  httpport: 9222,
}

nats: {
  Servers: ["nats://cog-nats.nats.svc.cluster.local:4222"],
  ConnectTimeout: 5000,
  MaxReconnects: 120,
  ReconnectWait: 5000,
}


connect: [
  {
    id: "IRISNATs"
    type: "NATSToKafka",
    subject: "iris.stream",
    topic: "iris-requests",
    brokers: ["kafka-cluster-kafka-bootstrap.kafka:9092"],
  },
  {
    type: "JetStreamToKafka",
    brokers: ["kafka-cluster-kafka-bootstrap.kafka:9092"]
    id: "irisstream-1",
    topic: "iris-requests-1",
    subject: "iris.stream",
    durablename: "durable_iris_consumer",
    queuename: "iris_consumers"
  },
]
"""



"""
This module is to test functionalities of kubeflowplugin.
"""

import os
import unittest
from unittest.mock import MagicMock, patch
from kubernetes.client.models import V1EnvVar
from ..cogflow import (
    pipeline,
    create_component_from_func,
    client,
    serve_model_v2,
    serve_model_v1,
    load_component,
    delete_pipeline,
)
from ..cogflow.plugins.kubeflowplugin import KubeflowPlugin, CogContainer


class TestKubeflowPlugin(unittest.TestCase):
    """Test cases for KubeflowPlugin class."""

    def setUp(self):
        """Set up method to initialize plugin."""
        self.kfp_plugin = KubeflowPlugin()

    @patch("kfp.dsl.pipeline")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_pipeline_with_name_and_description(
        self, mock_plugin_activation, mock_pipeline
    ):
        """Test pipeline function with name and description."""
        pipeline()
        mock_pipeline.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.components.create_component_from_func")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_create_component_from_func(
        self, mock_plugin_activation, mock_create_component
    ):
        """Test create_component_from_func."""
        func = MagicMock()
        create_component_from_func(func)
        mock_create_component.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.Client")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_client(self, mock_plugin_activation, mock_client):
        """Test client function."""
        # Arrange
        client()

        # Assertion
        mock_client.assert_called_once()
        mock_plugin_activation.assert_called_once()

    @patch("kfp.components.load_component_from_url")
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_load_component_from_url_success(
        self, mock_plugin_activation, mock_load_component
    ):
        """Test loading component from URL."""
        # Mock a successful component loading
        expected_component_spec = MagicMock()
        mock_load_component.return_value = expected_component_spec

        # Define a sample URL
        url = "http://example.com/component.tar.gz"

        # Call the function under test
        result = load_component(url=url)

        # Assert that the function returns the expected component specification
        self.assertEqual(result, expected_component_spec)

        # Assert that load_component_from_url was called with the correct URL
        mock_load_component.assert_called_once_with(url)
        mock_plugin_activation.assert_called_once()

    @patch.dict(
        os.environ,
        {
            "DB_HOST": "postgres",
            "DB_PORT": "5432",
            "DB_USER": "hiro",
            "DB_PASSWORD": "hiropwd",
            "DB_NAME": "cognitiveDB",
            "AWS_ACCESS_KEY_ID": "access_key_id_value",
            "AWS_SECRET_ACCESS_KEY": "secret_access_key_value",
            "MINIO_BUCKET_NAME": "MLFLOW",
            "MLFLOW_TRACKING_URI": "http://127.0.0.1:5001",
            "MINIO_ENDPOINT_URL": "localhost:9000",
            "MLFLOW_S3_ENDPOINT_URL": "C:/path/to/mlruns",
        },
    )
    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_add_model_access_with_valid_env_vars(self, mock_plugin_activation):
        """Test that valid environment variables are added to the container."""

        container = CogContainer()
        container_with_env_vars = container.add_model_access()

        # List of expected environment variables
        expected_env_vars = [
            V1EnvVar(name="DB_HOST", value="postgres"),
            V1EnvVar(name="DB_PORT", value="5432"),
            V1EnvVar(name="DB_USER", value="hiro"),
            V1EnvVar(name="DB_PASSWORD", value="hiropwd"),
            V1EnvVar(name="DB_NAME", value="cognitiveDB"),
            V1EnvVar(name="AWS_ACCESS_KEY_ID", value="access_key_id_value"),
            V1EnvVar(name="AWS_SECRET_ACCESS_KEY", value="secret_access_key_value"),
            V1EnvVar(name="MINIO_BUCKET_NAME", value="MLFLOW"),
            V1EnvVar(name="MLFLOW_TRACKING_URI", value="http://127.0.0.1:5001"),
            V1EnvVar(name="MINIO_ENDPOINT_URL", value="localhost:9000"),
            V1EnvVar(name="MLFLOW_S3_ENDPOINT_URL", value="C:/path/to/mlruns"),
        ]

        # Check that all expected environment variables are in the container
        actual_env_vars = [env_var.to_dict() for env_var in container_with_env_vars.env]
        mock_plugin_activation.assert_called_once()
        for expected_env_var in expected_env_vars:
            self.assertIn(expected_env_var.to_dict(), actual_env_vars)

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_serve_model_v2(self, mock_plugin_activation):
        """Test serving model v2."""
        # Patch Kubernetes client to avoid loading kube config
        with patch("kubernetes.config.load_kube_config"):
            model_uri = "sample_model_uri"
            name = "test_model_name"

            # Call the function and assert that it raises MaxRetryError
            with self.assertRaises(Exception):
                serve_model_v2(model_uri, name)
            mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_serve_model_v2_no_name(self, mock_plugin_activation):
        """Test serving model v2 without a name."""
        with patch("kubernetes.config.load_kube_config"):
            model_uri = "sample_model_uri"

            # Call the function and assert that it raises MaxRetryError
            with self.assertRaises(Exception):
                self.kfp_plugin.serve_model_v2(model_uri)
            mock_plugin_activation.assert_called_once()

    @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    def test_serve_model_v1_with_exception(self, mock_plugin_activation):
        """Test serving model v1 with an exception."""
        # Define input parameters
        model_uri = "example_model_uri"
        name = "test_model_name"

        # Call the function and assert that it raises MaxRetryError
        with self.assertRaises(Exception):
            serve_model_v1(model_uri, name)
        mock_plugin_activation.assert_called_once()

    # @patch("cogflow.cogflow.pluginmanager.PluginManager.verify_activation")
    # def test_get_model_url(self, mock_plugin_activation):
    #     """Test get model URL functionality."""
    #     model_name = "test_model"
    #
    #     with self.assertRaises(Exception):
    #         # Call the method you're testing here
    #         get_model_url(model_name)
    #     mock_plugin_activation.assert_called_once()

    @patch("requests.delete")
    @patch("requests.get")
    @patch("os.getenv")
    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_delete_pipeline(
        self, mock_client, mock_env, mock_request_get, mock_request_delete
    ):
        """Test deleting a pipeline."""
        # Arrange
        mock_env.side_effect = lambda x: {
            "MLFLOW_S3_ENDPOINT_URL": "localhost:9000",
            "AWS_ACCESS_KEY_ID": "minio",
            "AWS_SECRET_ACCESS_KEY": "minio123",
            "API_BASEPATH": "http://randomn",
            "TIMER_IN_SEC": "10",
            "FILE_TYPE": "2",
            "MLFLOW_TRACKING_URI": "http://mlflow",
            "ML_TOOL": "ml_flow",
            "COGFLOW_CONFIG_FILE_PATH": "/path/to/config",
        }[x]
        mock_request_get.return_value.status_code = 200
        mock_request_delete.return_value.status_code = 200
        mock_client_instance = mock_client.return_value
        pipeline_id = "test_pipeline_id"

        # Act
        delete_pipeline(pipeline_id)

        # Assert
        mock_client_instance.delete_pipeline.assert_called_once_with(
            pipeline_id=pipeline_id
        )

    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_list_pipeline_versions(self, mock_client):
        """Test listing pipeline versions."""
        # Arrange
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        pipeline_id = "test_pipeline_id"
        expected_response = "expected_response"
        mock_client_instance.list_pipeline_versions.return_value = expected_response

        # Act
        response = plugin.list_pipeline_versions(pipeline_id)

        # Assert
        mock_client_instance.list_pipeline_versions.assert_called_once_with(
            pipeline_id=pipeline_id
        )
        self.assertEqual(response, expected_response)

    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_delete_pipeline_version(self, mock_client):
        """Test deleting a pipeline version."""
        # Arrange
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        version_id = "test_version_id"

        # Act
        plugin.delete_pipeline_version(version_id)

        # Assert
        mock_client_instance.delete_pipeline_version.assert_called_once_with(
            version_id=version_id
        )

    @patch("cogflow.cogflow.plugins.kubeflowplugin.KubeflowPlugin.client")
    def test_delete_runs(self, mock_client):
        """Test deleting pipeline runs."""
        # Arrange
        plugin = KubeflowPlugin()
        mock_client_instance = mock_client.return_value
        mock_client_instance.runs = MagicMock()
        run_ids = [1, 2]

        # Act
        plugin.delete_runs(run_ids)

        # Assert
        calls = [unittest.mock.call(id=1), unittest.mock.call(id=2)]
        mock_client_instance.runs.delete_run.assert_has_calls(calls, any_order=True)
        self.assertEqual(mock_client_instance.runs.delete_run.call_count, 2)

    @patch("kubernetes.config.load_kube_config")
    def test_outside_cluster_with_mock_kubeconfig(self, mock_load_kube_config):
        """Test kubeconfig outside cluster"""
        # Mock the kubeconfig return value (example structure)
        mock_load_kube_config.return_value = {
            "clusters": [
                {"name": "cluster-1", "cluster": {"server": "http://example.com"}}
            ],
            "contexts": [{"name": "context-1", "context": {"namespace": "default"}}],
            "users": [{"name": "user-1", "user": {"token": "fake-token"}}],
        }

        # Act: Now test the function with the mocked kubeconfig
        kubeflow_plugin = KubeflowPlugin()
        result = kubeflow_plugin.get_default_namespace()

        # Assert: Check the namespace in the context
        assert result == "default", f"Expected 'development', but got {result}"


if __name__ == "__main__":
    unittest.main()
